"""

test device

"""
import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import threading
import copy
from redvypr.device import RedvyprDeviceCustomConfig, RedvyprDevice
import redvypr.data_packets
from redvypr.data_packets import check_for_command
import pydantic

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('test_device')
logger.setLevel(logging.DEBUG)

redvypr_devicemodule = True
class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = False
    description: str = 'A simple test device'

class DeviceCustomConfig(RedvyprDeviceCustomConfig):
    delay_s: float = 1.0

def start(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start():'
    logger.debug(funcname)
    print('Config',config)
    pdconfig = DeviceCustomConfig.model_validate(config)
    print('pdconfig',pdconfig)
    #data = {'_keyinfo':config['_keyinfo']}
    # dataqueue.put(data)
    # Send a datapacket with metadata describing the device
    address_str = device_info['address_str']
    device_metadata = {'location':'Room 42'}
    datapacket_info_device = redvypr.data_packets.add_metadata2datapacket(datapacket={}, address=address_str, metadict=device_metadata)
    dataqueue.put(datapacket_info_device)
    # Send a datapacket with information once (that will be put into the statistics)
    datapacket_info = redvypr.data_packets.add_metadata2datapacket(datapacket={}, datakey='sine_rand', metakey='unit',metadata='random unit')
    # Metadata can also be given as a dict
    metadata = {'description':'sinus with random data', 'mac':'ABCDEF1234'}
    datapacket_info = redvypr.data_packets.add_metadata2datapacket(datapacket_info, datakey='sine_rand', metadict=metadata)
    dataqueue.put(datapacket_info)
    i = 0
    counter = 0
    while True:
        try:
            data = datainqueue.get(block = False)
        except:
            data = None
        if(data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            logger.debug('Got a command: {:s}'.format(str(data)))
            if (command == 'stop'):
                logger.debug('Command is for me: {:s}'.format(str(command)))
                break

        data = redvypr.data_packets.create_datadict(device = device_info['device'])
        data['data'] = float(np.random.rand(1)-0.5)
        data['sometext'] = 'Hello {}'.format(counter)
        dataqueue.put(data)
        # Calculate some sine
        data_rand = float(np.random.rand(1) - 0.5)
        data_rand_2 = float(np.random.rand(1) - 0.5)
        f_sin = 1 / 30  # Frequency in Hz
        A_sin = 10  # Amplitude
        data_sine = float(A_sin * np.sin(f_sin * time.time()))
        dataqueue.put({'sine_rand': data_rand + data_sine,'count': counter,'sine': data_sine})

        time.sleep(config['delay_s'])
        #print('Hallo')
        # Add complex data
        data = redvypr.data_packets.create_datadict(device='test_complex_data', packetid='complex_data')
        if counter == 0:
            # Add metadata

            metadata = {'unit': 'baseunit'}
            data = redvypr.data_packets.add_metadata2datapacket(data, datakey='data_list_list',
                                                                metadict=metadata)

            metadata = {'unit': 'otherunit of entry 0'}
            data = redvypr.data_packets.add_metadata2datapacket(data, datakey='["data_list_list"][0]',
                                                                metadict=metadata)

            metadata = {'description': 'Counter and polynomial functions of counter', 'unit': 'grigra'}
            data = redvypr.data_packets.add_metadata2datapacket(data, datakey='data_list_poly',
                                                                           metadict=metadata)

            metadata = {'description': 'Temperature', 'unit': 'degC'}
            data = redvypr.data_packets.add_metadata2datapacket(data, datakey='["data_dict_list"]["temp"]',
                                                                metadict=metadata)

            metadata = {'unit': 'Pa'}
            data = redvypr.data_packets.add_metadata2datapacket(data, datakey='["data_dict_list"]["pressure"]',
                                                                metadict=metadata)
        data['data_list'] = [counter,data_sine,data_rand]
        data['data_list_list'] = [[counter, data_sine, data_rand],[counter, data_sine]]
        data['data_list_poly'] = [counter, counter + data_rand, 2 * counter + data_rand, -10 * counter + data_rand+ 3, 0.1 * counter**2 + 2 * counter + data_rand+ 3]
        data['data_dict_list'] = {'temp':[data_rand, 2*data_rand-10],'pressure':10+data_rand}
        data['data_ndarray_1d'] = np.zeros((5,)) + counter
        data['data_ndarray_2d'] = np.zeros((6,7)) + counter
        data['data_ndarray_2d_int'] = np.zeros((3,2),dtype=int) + int(counter)
        #print('datastreams',redvypr.data_packets.Datapacket(data).datastreams(expand=True))
        dataqueue.put(data)
        # Put some pathological data into the queue
        dataqueue.put(None)
        counter += 1
        






import datetime
import numpy as np
import logging
import sys
import threading
import copy
import yaml
import json
import typing
import pyqtgraph
import pydantic
import numpy
from PyQt5 import QtWidgets, QtCore, QtGui

import redvypr
from redvypr.device import RedvyprDevice, device_start_standard
from redvypr.data_packets import check_for_command
#from redvypr.packet_statistics import get_keys_from_data
#import redvypr.packet_statistic as redvypr_packet_statistic
import redvypr.data_packets as data_packets
import redvypr.gui as gui
#import redvypr.config as redvypr_config
from redvypr.redvypr_address import RedvyprAddress
#from redvypr.devices.plot import plot_widgets
from redvypr.devices.plot import XYplotWidget
import redvypr.files as redvypr_files

_icon_file = redvypr_files.icon_file

redvypr_devicemodule = True

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('pcolorplot')
logger.setLevel(logging.INFO)

start = device_start_standard


class configPcolorPlot(pydantic.BaseModel):
    location: list  = pydantic.Field(default=[])
    type: str = 'PcolorPlot'
    dt_update: float = pydantic.Field(default=0.25,description='Update time of the plot [s]')
    datastream: RedvyprAddress = pydantic.Field(default=RedvyprAddress(''))

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = False
    subscribes: bool = True
    description: str = 'Device to plot Pcolor-Data'
    gui_icon: str = 'ph.chart-line-fill'


#class DeviceCustomConfig(configPcolorPlot):
class DeviceCustomConfig(pydantic.BaseModel):
    location: list  = pydantic.Field(default=[])
    type: str = 'PcolorPlot'
    dt_update: float = pydantic.Field(default=0.25, description='Update time of the plot [s]')
    datastream: RedvyprAddress = pydantic.Field(default=RedvyprAddress(''))

# Use the standard start function as the start function

class PcolorPlotWidget(QtWidgets.QWidget):
    def __init__(self, config=None, redvypr_device=None, loglevel=logging.DEBUG):
        funcname = __name__ + '.init():'
        super(QtWidgets.QWidget, self).__init__()
        self.device = redvypr_device
        if self.device is not None:
            self.redvypr = self.device.redvypr
        else:
            self.redvypr = None

        if (config == None):  # Create a config from the template
            self.config = configPcolorPlot()
        else:
            self.config = config
        self.logger = logging.getLogger('XYplot')
        self.logger.setLevel(loglevel)
        self.description = 'Pcolor plot'
        self.layout = QtWidgets.QGridLayout(self)
        self.data_z = []
        self.data_x = []
        self.data_y = []
        self.data_all = []
        self.create_widgets()

    def create_widgets(self):

        self.graphiclayout = pyqtgraph.GraphicsLayoutWidget()
        self.plotwidget = self.graphiclayout.addPlot()
        z = numpy.random.rand(10,10)
        pcmi = pyqtgraph.PColorMeshItem()
        pcmi.setData(z)
        colbar = pyqtgraph.ColorBarItem()
        colbar.setImageItem(pcmi)
        self.mesh = pcmi
        self.plotwidget.addItem(pcmi)
        self.graphiclayout.addItem(colbar)
        axis = pyqtgraph.DateAxisItem(orientation='bottom', utcOffset=0)
        self.plotwidget.setAxisItems({"bottom": axis})
        self.layout.addWidget(self.graphiclayout, 0, 0, 1, 2)


    def update_data(self,rdata):
        funcname = __name__ + '.update_data():'
        self.logger.debug(funcname + 'Got data {}'.format(rdata))
        data_plot = rdata[self.config.datastream]
        print('Data plot', data_plot)
        self.data_z.append(data_plot)
        self.data_all.append(rdata)
        self.data_x.append(rdata['t'])
        if len(self.data_x) > 2:
            try:
                z = numpy.asarray(self.data_z)
                Z = z[:-1, :]
                ny = numpy.shape(z)[1]
                nx = numpy.shape(z)[0]
                y = numpy.arange(0,ny+1)
                x = numpy.asarray(self.data_x)
                X = numpy.asarray(numpy.tile(x,(ny+1,1)))
                X = X.T
                Y = numpy.asarray(numpy.tile(y, (nx, 1)))
                #Y = Y.T
                print('x',x)
                print('X', X)
                print('y', y)
                print('Y', Y)
                print('z', z)
                print('shapes', numpy.shape(X),numpy.shape(Y),numpy.shape(Z))
                self.mesh.setData(X,Y,Z)
            except:
                logger.warning('Could not update data',exc_info=True)

class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None,tabwidget=None):
        funcname = __name__ + '__init__():'
        logger.debug(funcname)
        super(QtWidgets.QWidget, self).__init__()
        self.device = device
        self.layout = QtWidgets.QGridLayout(self)
        self.pcolorplot = PcolorPlotWidget(redvypr_device=device, config=self.device.custom_config)
        self.layout.addWidget(self.pcolorplot)
        self.device.config_changed_signal.connect(self.config_changed)

    def config_changed(self):
        print('XYplot config changed')
        print('Config',self.device.custom_config)
        # Check if subscriptions need to be changed
        self.pcolorplot.config = self.device.custom_config
        #self.pcolorplot.apply_config()

    def update_data(self, data):
        funcname = __name__ + '.update_data():'
        try:
            print(funcname)
            print('Got data', data)
            print('Datastream', self.device.custom_config.datastream)
            #self.pcolorplot.update_plot(data)
            rdata = redvypr.data_packets.Datapacket(data)
            if rdata in self.device.custom_config.datastream:
                self.pcolorplot.update_data(rdata)

        except:
            pass
            #logger.warning(funcname + 'Could not process data',exc_info=True)


import redvypr
import time

raddr = redvypr.RedvyprAddress('*')
raddr_index = redvypr.RedvyprAddress("/k:['somelist'][3]")
datapacket_raw = {'t1':3,'somestring':'Hello World','somelist':[1,2,5,7],'somedict':{'test':'ok','number':3.0}}
datapacket = datapacket_raw.copy()
# Create a bogus redvypr host
print('Creating hostinfo')
hostinfo = redvypr.create_hostinfo(hostname='someredvypr')
devicename = 'somedevice'
devicemodulename = 'somedevicemodulename'
numpacket = 0
tread = time.time() # The time the packet was received from the queue (in redvypr.distribute_data)
redvypr.packet_statistic.treat_datadict(datapacket, devicename, hostinfo, numpacket, tread,devicemodulename)
print('Datapacket', datapacket)

#
print('Retrieving data')
print('---------------')
print('---------------')
rdatapacket = redvypr.data_packets.Datapacket(datapacket)
print('With a redvypr datapacket (basically an improved dictionary) data can be retrieved')
print('1:\tUsing the keyword functionality of dictionaries')
print('''data = rdatapacket['t1']''')
data = rdatapacket['t1']
print('data',data)
print('---------------')
print('2:\tUsing the keyword functionality of dictionaries but with square brackets to allow to access members of lists etc')
print('This is realized using the eval functionality ...\n')
print('''data = rdatapacket["['somedict']['test']"]''')
data = rdatapacket["['somedict']['test']"]
print('data',data)
print('---------------')
print('3:\tUsing a redvypr address\n')
print('''raddr = redvypr.RedvyprAddress("/k:['somelist'][2]")''')
print('''data = rdatapacket[raddr]''')
raddr = redvypr.RedvyprAddress("/k:['somelist'][2]")
data = rdatapacket[raddr]
print('data',data)
print('---------------')
#print('address',raddr_index)
#print('R datapacket',rdatapacket)





import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import yaml
import copy
import pyqtgraph

import redvypr.data_packets
import redvypr.gui
import redvypr.files as files
#from redvypr.configdata import configtemplate_to_dict, configdata, getdata
from copy import deepcopy as dc

_logo_file = files.logo_file
_icon_file = files.icon_file
pyqtgraph.setConfigOption('background', 'w')
pyqtgraph.setConfigOption('foreground', 'k')

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('plot_widgets')
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)



#
#                
#
# Plot widget
#
#
#

config_template_grid_loc = {}
config_template_grid_loc['template_name'] = 'gridloc'
config_template_grid_loc['x'] = {'type': 'int', 'default': 0}
config_template_grid_loc['y'] = {'type': 'int', 'default': 0}
config_template_grid_loc['width'] = {'type': 'int', 'default': 1}
config_template_grid_loc['height'] = {'type': 'int', 'default': 1}

description_graph = 'Device that plots the received data'
config_template_graph_line = {}
config_template_graph_line['template_name'] = 'Line'
config_template_graph_line['buffersize'] = {'type': 'int', 'default': 2000,
                                           'description': 'The size of the buffer holding the data of the line'}
config_template_graph_line['name'] = {'type': 'str', 'default': '$y',
                                     'description': 'The name of the line, this is shown in the legend'}
config_template_graph_line['x'] = {'type': 'datastream', 'default': '$t(y)',
                                  'description': 'The x-data of the plot'}
config_template_graph_line['y'] = {'type': 'datastream', 'default': '',
                                  'description': 'The y-data of the plot'}
config_template_graph_line['color'] = {'type': 'color', 'description': 'The color of the plot'}
config_template_graph_line['linewidth'] = {'type': 'int', 'default': 1,
                                          'description': 'The linewidth of the line'}
config_template_graph = {}
config_template_graph['template_name'] = 'Realtime graph'
config_template_graph['location']    = config_template_grid_loc
config_template_graph['type']        = {'type': 'str', 'default': 'graph', 'modify': False}
config_template_graph['dt_update']   = {'type': 'float', 'default': 0.25, 'modify': True,'description':'update time of plot'}
config_template_graph['interactive'] = {'type': 'str', 'default': 'mouse', 'modify': True,'description':'Interactive modes'}
config_template_graph['backgroundcolor'] = {'type': 'color', 'default': 'lightgray'}
config_template_graph['bordercolor'] = {'type': 'color', 'default': 'lightgray'}

config_template_graph['useprops'] = {'type': 'bool', 'default': True,
    'description': 'Use the properties to display units etc.'}

config_template_graph['showunits'] = {'type': 'bool', 'default': True,
    'description': 'Add the unit of the y-data to the legend, queried from the datakey.'}
config_template_graph['datetick'] = {'type': 'bool', 'default': True,
    'description': 'x-axis is a date axis'}

config_template_graph['title'] = {'type': 'str', 'default': ''}
config_template_graph['xlabel'] = {'type': 'str', 'default': ''}
config_template_graph['ylabel'] = {'type': 'str', 'default': ''}
config_template_graph['lines'] = {'type': 'list', 'default': [config_template_graph_line], 'modify': True,
                                 'options': [config_template_graph_line]}
#config_template_graph['description'] = description_graph



class redvypr_graph_widget(QtWidgets.QFrame):
    """ Widget is plotting realtimedata using the pyqtgraph functionality
    This widget can be configured with a configuration dictionary 
    """
    def __init__(self,config=None, redvypr_device = None , loglevel = logging.DEBUG):
        funcname = __name__ + '.init():'
        super(QtWidgets.QFrame, self).__init__()
        self.device = redvypr_device
        self.logger = logging.getLogger('plot_graph_widget')
        self.logger.setLevel(loglevel)
        self.description = description_graph
        self.config_template = config_template_graph

        if(config == None): # Create a config from the template
            self.config = redvypr.config.configuration(self.config_template)
        else:
            self.config = redvypr.config.configuration(self.config_template,config=config)

        #self.logger.debug('plot widget config {:s}'.format(str(self.config)))
        backcolor = str(self.config['backgroundcolor'])
        bordercolor = str(self.config['bordercolor'])
        #logger.debug(funcname + 'backcolor {:s}'.format(str(backcolor)))
        style = "background-color : {:s};border : 1px solid {:s};".format(backcolor, bordercolor)
        #logger.debug(funcname + 'Style: {:s}'.format(str(style)))
        self.setStyleSheet(style)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.create_widgets()
        self.apply_config()


    def create_widgets(self):


        """
        Creates the configuration

        Returns:

        """
        funcname = __name__ + '.create_widgets()'
        config = self.config
        logger.debug(funcname)
        i = 0
        if True:
            logger.debug(funcname + ': Adding plot ')# + str(config))
            try:
                title = config['title'].data
            except:
                title = "Plot {:d}".format(i)
                
            try:
                name = config['name'].data
            except:
                name = "Plot {:d}".format(i)

            # https://stackoverflow.com/questions/44402399/how-to-disable-the-default-context-menu-of-pyqtgraph
            plot = pyqtgraph.PlotWidget(title=title,name=name)
            menu = plot.plotItem.vb.menu.addMenu('Redvypr config')
            action = menu.addAction('Line config')
            action.triggered.connect(self.pyqtgraphRedvyprAction)

            #plot = redvyprPlotWidget(title=title,name=name)
            plot.register(name=name)
            # Add a legend
            legend = plot.addLegend()
            if config['interactive'].lower() == 'mouse':
                self.logger.debug('Adding interactive mouse (vertical line)')
                plot.scene().sigMouseMoved.connect(self.mouseMoved)
                plot.scene().sigMouseClicked.connect(self.mouseClicked)
                self.vb = plot.plotItem.vb
                # Add a vertical line
                color = QtGui.QColor(200,100,100)
                linewidth = 2
                pen = pyqtgraph.mkPen(color, width=linewidth)
                self.vLineMouse = pyqtgraph.InfiniteLine(angle=90, movable=False, pen=pen)
                #print('Set pen 3')
                plot.addItem(self.vLineMouse, ignoreBounds=True)

            self.layout.addWidget(plot)
                
            config.plot = plot
            config.legend = legend
            #plot_dict = {'widget': plot, 'lines': []}

                
                
        #self.plot_dict = plot_dict


    def pyqtgraphRedvyprAction(self,hallo):
        """

        """
        funcname = __name__ + '.pyqtgraphRedvyprAction()'
        logger.debug(funcname)
        self.configWidget = redvypr.gui.configWidget(self.config)
        self.configWidget.setWindowIcon(QtGui.QIcon(_icon_file))
        self.configWidget.config_changed_flag.connect(self.apply_config)
        self.configWidget.show()


    def mouseMoved(self,evt):
        """Function if mouse has been moved
        """
        pos = (evt.x(),evt.y())
        mousePoint = self.vb.mapSceneToView(evt)
        #print('mouse moved',mousePoint.x())
        self.vLineMouse.setPos(mousePoint.x())
        #for vline in self.vlines:
        #    vline.setPos(mousePoint.x())

    def mouseClicked(self,evt):
        #col =
        #col = pg.mkPen(0.5,width=3)
        #colsymbol = pg.mkPen(color=QtGui.QColor(150,150,150),width=4)
        #print('Clicked: ' + str(evt.scenePos()))
        pass

    def set_title(self,title):
        """
        Sets the title of the plot
        """
        funcname = __name__ + '.set_title()'
        logger.debug(funcname)
        self.config['title'].data = title
        self.config.plot.setTitle(title)

    def set_line(self, index, y, name='$y', x='$t(y)', color=[255, 0, 0], linewidth=1, bufsize=2000):
        """
        Sets the line at index to the parameters
        """

        self.add_line(y, name, x, color, linewidth, bufsize, index = index)

    def add_line(self,y,name='$y',x='$t(y)', color=[255,0,0], linewidth=1, bufsize=2000, index = None):
        """
        Adds a line to the plot
        """
        funcname = __name__ + '.add_line()'
        logger.debug(funcname)
        if index is None:
            rconfig = redvypr.config.configuration(config_template_graph_line)
            self.config['lines'].data.append(rconfig)
            index = len(self.config['lines']) - 1

        config_line = {}
        self.config['lines'][index]['buffersize'].data = bufsize
        self.config['lines'][index]['linewidth'].data = linewidth
        qcolor_tmp = redvypr.gui.get_QColor(color)
        self.config['lines'][index]['color']['r'].data = qcolor_tmp.red()
        self.config['lines'][index]['color']['g'].data = qcolor_tmp.green()
        self.config['lines'][index]['color']['b'].data = qcolor_tmp.blue()
        self.config['lines'][index]['x'].data = x
        self.config['lines'][index]['y'].data = y
        self.config['lines'][index]['name'].data = name

        self.apply_config()

        #config_template_graph_line = {}
        #config_template_graph_line['template_name'] = 'Line'
        #config_template_graph_line['buffersize'] = {'type': 'int', 'default': 2000,
        #                                            'description': 'The size of the buffer holding the data of the line'}
        #config_template_graph_line['name'] = {'type': 'str', 'default': '$y',
        #                                      'description': 'The name of the line, this is shown in the legend'}
        #config_template_graph_line['x'] = {'type': 'datastream', 'default': '$t(y)',
        #                                   'description': 'The x-data of the plot'}
        #config_template_graph_line['y'] = {'type': 'datastream', 'default': '',
        #                                   'description': 'The y-data of the plot'}
        #config_template_graph_line['color'] = {'type': 'color', 'description': 'The color of the plot'}
        #config_template_graph_line['linewidth'] = {'type': 'int', 'default': 1,
        #                                           'description': 'The linewidth of the line'}

    def apply_config(self):
        """
        Function is called by the initialization or after the configuration was changed

        Returns:

        """
        funcname = __name__ + '.apply_config()'
        logger.debug(funcname)
        plot = self.config.plot
        # Title
        title = self.config['title'].data
        plot.setTitle(title)

        try:
            datetick = self.config['datetick'].data
        except:
            datetick = False

        if datetick:
            logger.debug(funcname + 'Datetick')
            axis = pyqtgraph.DateAxisItem(orientation='bottom',utcOffset=0)
            plot.setAxisItems({"bottom": axis})
        else:
            logger.debug(funcname + ' No Datetick')
            axis = pyqtgraph.AxisItem(orientation='bottom')
            plot.setAxisItems({"bottom": axis})


        # Label
        # If a xlabel is defined
        try:
            plot.setLabel('left', self.config['ylabel'].data)
        except:
            pass

        # If a ylabel is defined
        try:
            plot.setLabel('bottom', self.config['xlabel'].data)
        except:
            pass

        # Add lines with the actual data to the graph
        for iline, line in enumerate(self.config['lines']):
            #print('Line',line)
            #FLAG_HAVE_LINE = False

            # check if we have already a lineplot, if yes, dont bother
            try:
                line.lineplot
                continue
            except:
                pass

            logger.debug(funcname + ': Adding a line to the plot:')# + str(line))
            buffersize = line['buffersize'].data
            tdata = np.zeros(buffersize) * np.NaN
            xdata = np.zeros(buffersize) * np.NaN
            ydata = np.zeros(buffersize) * np.NaN

            try:
                name = line['name'].data
            except:
                name = 'line {:d}'.format(iline)

            lineplot = pyqtgraph.PlotDataItem(name=name)
            line.lineplot = lineplot # Add the line as an attribute to the configuration
            try:
                x = line['x'].data
            except:
                pass

            try:
                y = line['y'].data
            except:
                pass

            try:
                linewidth = line['linewidth'].data
            except:
                linewidth = 1

            try:
                color = redvypr.gui.get_QColor(line['color'])
            except Exception as e:
                logger.debug('Definition of color not usable:')
                logger.exception(e)
                color = QtGui.QColor(255, 10, 10)

            # Configuration of the line plot
            lineconfig = {'x': x, 'y': y, 'linewidth': linewidth, 'color': color}
            # Add the line and the configuration to the lines list
            line_dict = {'line': lineplot, 'config': lineconfig, 'xdata': xdata, 'ydata': ydata, 'tdata':tdata,'tlastupdate':0}
            line.line_dict = line_dict
            plot.addItem(lineplot)
            # Configuration


        # Update the line configuration
        self.config.legend.clear()
        for iline, line in enumerate(self.config['lines']):
            try:
                logger.debug(funcname + ': Updating line {:d}'.format(iline))
                #print('Line',line,iline)

                lineconfig = line.line_dict['config']
                #print('Lineconfig',lineconfig)
                # The data buffer
                tdata = line.line_dict['tdata']
                xdata = line.line_dict['xdata']
                ydata = line.line_dict['ydata']
                if(len(xdata) != line['buffersize']):
                    buffersize = line['buffersize']
                    print('Updating the buffersize')
                    line.line_dict['tdata'] = np.zeros(buffersize) * np.NaN
                    line.line_dict['xdata'] = np.zeros(buffersize) * np.NaN
                    line.line_dict['ydata'] = np.zeros(buffersize) * np.NaN


                x = line['x'].data
                y = line['y'].data
                yaddr = redvypr.data_packets.redvypr_address(y)
                if(x == '$t(y)'):
                    logger.debug(funcname + ' Using time variable of y')
                    xtmp = redvypr.data_packets.modify_addrstr(yaddr.address_str,datakey='t')
                    #print('xtmp',xtmp)
                    xaddr = redvypr.data_packets.redvypraddress(xtmp)
                else:
                    xaddr = redvypr.data_packets.redvypraddress(x)

                # These attributes are used in plot.Device.connect_devices to actually subscribe to the fitting devices
                line['x'].xaddr = xaddr
                line['y'].yaddr = yaddr
                lineconfig['x'] = x
                lineconfig['y'] = y
                lineconfig['xaddr'] = xaddr
                lineconfig['yaddr'] = yaddr

                #print('Set pen')
                lineplot = line.line_dict['line']  # The line to plot
                #print('Set pen 1')
                color = redvypr.gui.get_QColor(self.config['lines'][iline]['color'])
                #print('COLOR!!!!',color)
                #print('Set pen 2')
                linewidth = self.config['lines'][iline]['linewidth'].data
                #print('Set pen 3')
                pen = pyqtgraph.mkPen(color, width=linewidth)
                lineplot.setPen(pen)
                name = self.config['lines'][iline]['name'].data
                if(name.lower() == '$y'):
                    logger.debug(funcname + ' Replacing with y')
                    name = y

                logger.debug(funcname + ' Setting the name')
                self.config.legend.addItem(lineplot,name)
                lineplot.setData(name=name)
            except Exception as e:
                logger.exception(e)
                logger.debug('Exception config lines: {:s}'.format(str(e)))


        logger.debug(funcname  + ' done.')

        
    def clear_buffer(self):
        """ Clears the buffer of all lines
        """
        # Check if the device is to be plotted
        
        for iline, line in enumerate(self.config['lines']):
            line_dict = line.line_dict
            line      = line_dict['line'] # The line to plot
            config    = line_dict['config'] # The line to plot
            line_dict['tdata'][:] = np.NaN
            line_dict['xdata'][:] = np.NaN
            line_dict['ydata'][:] = np.NaN
            # Set the data
            x = line_dict['xdata']  # The line to plot
            y = line_dict['ydata']  # The line to plot
            line.setData(x=x, y=y)

    def get_data(self, xlim):
        """
        Gets the data of the buffer in the limits of xlim
        """
        funcname = __name__ + '.get_data():'
        tdata = []
        xdata = []
        ydata = []
        for iline, line in enumerate(self.config['lines']):
            line_dict = line.line_dict
            line = line_dict['line']  # The line to plot
            config = line_dict['config']  # The line to plot
            t = line_dict['tdata']  # The line to plot
            x = line_dict['xdata']  # The line to plot
            y = line_dict['ydata']  # The line to plot
            ind = (x > xlim[0]) & (x < xlim[1])
            tdata_tmp = t[ind]
            xdata_tmp = x[ind]
            ydata_tmp = y[ind]
            tdata.append(tdata_tmp)
            xdata.append(xdata_tmp)
            ydata.append(ydata_tmp)
            logger.debug(funcname + ' Got data of length {:d}'.format(len(tdata_tmp)))
            #print('get_data',datetime.datetime.utcfromtimestamp(tdata_tmp[0]),datetime.datetime.utcfromtimestamp(xdata_tmp[0]))


        return {'x':xdata,'y':ydata,'t':tdata}


        
    def update_plot(self, data):
        """ Updates the plot based on the given data
        """
        funcname = self.__class__.__name__ + '.update_plot():'
        tnow = time.time()
        #print(funcname + 'got data',data,tnow)


        #try:
        if True:
            # Loop over all plot axes
            if True:
                # Check if the device is to be plotted
                for iline, line in enumerate(self.config['lines']):
                    line_dict = line.line_dict
                    #print('line dict',line_dict)
                    xaddr = line_dict['config']['xaddr']
                    yaddr = line_dict['config']['yaddr']
                    #print('adresses:',xaddr.get_str(),yaddr.get_str())
                    #print('device',data['_redvypr']['device'])
                    #print('data in',(data in xaddr),(data in yaddr))
                    if (data in xaddr) and (data in yaddr):
                        pw        = self.config.plot # The plot widget
                        line      = line_dict['line'] # The line to plot
                        config    = line_dict['config'] # The line to plot
                        tdata     = line_dict['tdata']  # The line to plot
                        xdata     = line_dict['xdata'] # The line to plot
                        ydata     = line_dict['ydata'] # The line to plot
                        # data can be a single float or a list, if its a list add it item by item
                        newt = data['_redvypr']['t'] # Add also the time of the packet
                        newx = data[xaddr.datakey]
                        newy = data[yaddr.datakey]
                        if(type(newx) is not list):
                            newx = [newx]
                        if (type(newy) is not list):
                            newy = [newy]

                        if(len(newx) != len(newy)):
                            logger.warning('lengths of x and y data different (x:{:d}, y:{:d})'.format(len(newx),len(newy)))
                            return

                        for inew in range(len(newx)): # TODO this can be optimized using indices instead of a loop
                            tdata        = np.roll(tdata, -1)
                            xdata        = np.roll(xdata,-1)
                            ydata        = np.roll(ydata,-1)
                            tdata[-1]    = float(newt)
                            xdata[-1]    = float(newx[inew])
                            ydata[-1]    = float(newy[inew])
                            line_dict['tdata']  = tdata
                            line_dict['xdata']  = xdata
                            line_dict['ydata']  = ydata

                        # Show the unit in the legend, if wished by the user and we have access to the device that can give us the metainformation
                        if (self.config['showunits']) and (self.device is not None):
                            try:
                                line.datakeys # datakeys found, doing nothing
                            except:
                                line.datakeys = self.device.get_datakeyinfo(yaddr)
                                logger.debug(funcname + ' Datakeyinfo {:s}'.format(str(line.datakeys)))
                                unit = None
                                for k in line.datakeys.keys():
                                    #print('key',k,yaddr,k in yaddr)
                                    if k in yaddr:
                                        print('fsfsd')
                                        try:
                                            unit = line.datakeys[k]['unit']
                                            break
                                        except:
                                            unit = None

                                if unit is not None:
                                    name = line.name()
                                    name_new = name + ' [{:s}]'.format(unit)
                                    self.config['lines'][iline]['name'].data = name_new
                                    self.apply_config()

            # Update the lines plot
            for iline, line in enumerate(self.config['lines']):
                line_dict = line.line_dict
                tlastupdate = line_dict['tlastupdate']  # The time the plot was last updated
                # Check if an update of the plot shall be done, or if only the buffer is updated
                dt = tnow - tlastupdate
                if dt > self.config['dt_update']:
                    update = True
                    line_dict['tlastupdate'] = tnow
                    #print('update')
                else:
                    update = False
                    #print('no update')

                if (update):  # We could check here if data was changed above the for given line
                    line      = line_dict['line'] # The line to plot
                    config    = line_dict['config'] # The line to plot
                    x         = line_dict['xdata'] # The line to plot
                    y         = line_dict['ydata'] # The line to plot
                    line.setData(x=x,y=y)
                    #pw.setXRange(min(x[:ind]),max(x[:ind]))

        #except Exception as e:
        #    #print(funcname + 'Exception:' + str(e))
        #    logger.debug(funcname + 'Exception:' + str(e))



#
#                
#
# numeric display widget
#
#
#
description_numdisp = 'Device that plots the received data'
#rdvpraddr = redvypr.data_packets.redvypr_address('tmp')
config_template_numdisp = {}
config_template_numdisp['template_name'] = 'Numeric display'
config_template_numdisp['type'] = {'type': 'str', 'default': 'numdisp', 'modify': False}
config_template_numdisp['location'] = config_template_grid_loc
config_template_numdisp['backgroundcolor'] = {'type': 'color', 'default': 'lightgray'}
config_template_numdisp['bordercolor'] = {'type': 'color', 'default': 'lightgray'}
config_template_numdisp['fontsize'] = {'type': 'int', 'default': 20}
config_template_numdisp['datastream'] = {'type': 'datastream', 'default': 'NA'}
config_template_numdisp['timeformat'] = {'type': 'str', 'default': '%d-%b-%Y %H:%M:%S'}
config_template_numdisp['unit'] = {'type': 'str', 'default': ''}
config_template_numdisp['datastreamlabel'] = {'type': 'str', 'options': ['tmp1','tmp2'],
                                           'default': '/d:/k:', 'description': 'Display the datastreamlabel'}
config_template_numdisp['useprops'] = {'type': 'bool', 'default': True,
                                    'description': 'Use the properties to display units etc.'}
config_template_numdisp['dataformat'] = {'type': 'str', 'default': 'f',
                                      'description': 'The format the data is shown, this will be interpreted as "{:dataformat}".format(data)'}
config_template_numdisp['title'] = {'type': 'str', 'default': 'test'}
config_template_numdisp['description'] = description_numdisp

class redvypr_numdisp_widget(QtWidgets.QFrame):
    """ Widget is plotting realtimedata on a display
    """
    def __init__(self,config=None):
        funcname = __name__ + '.init()'
        super(QtWidgets.QFrame, self).__init__()
        self.redvypr_addrtmp = redvypr.data_packets.redvypr_address('tmp')
        #self.setStyleSheet("border : 1px solid lightgray;background-color : lightgray")
        #self.setStyleSheet("backgrounNAd-color : red")
        self.description = description_numdisp
        self.config_template = config_template_numdisp

        print('Creating configuration')
        if(config == None): # Create a config from the template
            self.config = redvypr.config.configuration(self.config_template)
        else:
            self.config = redvypr.config.configuration(self.config_template,config=config)

        print('widget config',self.config)

        print('tpye',type(self.config['title'].data))
        self.titledisp = QtWidgets.QLabel(self.config['title'].data)
        self.titledisp.hide()

        self.devicedisp = QtWidgets.QLabel(self.config['datastream'].data)
        self.devicedisp.hide()

        self.timedisp = QtWidgets.QLabel(self.get_timestr(0,format=self.config['timeformat'].data))
        self.timedisp.hide()

        self.unitdisp = QtWidgets.QLabel(self.config['unit'].data)
        self.unitdisp.hide()


        self.setStyleSheet("background-color : {:s};border : 1px solid {:s};".format(self.config['backgroundcolor'].data,self.config['bordercolor'].data))
        self.layout = QtWidgets.QGridLayout(self)
        if True:
            logger.debug(funcname + ': Adding plot' + str(self.config))
            self.apply_config()

            self.numdisp = QtWidgets.QLabel("#")
            self.numdisp.setStyleSheet("border : 1px solid {:s};font-weight: bold; font-size: {:d}pt".format(self.config['backgroundcolor'].data,self.config['fontsize'].data))
            #
            self.layout.addWidget(self.numdisp,3,0)


    def apply_config(self):
        """
        Applies the config to the widgets
        Returns:

        """
        funcname = __name__ + '.apply_config():'
        logger.debug(funcname)
        title = self.config['title'].data
        if (len(title) > 0): # Show title
            self.titledisp.setText(title)
            self.titledisp.setStyleSheet("font-weight: bold;border : 1px solid {:s};".format(self.config['backgroundcolor'].data))
            self.titledisp.setAlignment(QtCore.Qt.AlignCenter)
            self.layout.addWidget(self.titledisp, 0, 0, 1, 2)
            self.titledisp.show()
        else: # Do not show title
            try:
                self.layout.removeWidget(self.titledisp)
                self.titledisp.hide()
            except Exception as e:
                print('problem',e)
                pass

        if (getdata(self.config['datastreamlabel'])): # TODO, let the user choose the datastream display options (UUID, key, key + address ...)
            self.redvypr_addrconv = redvypr.data_packets.redvypraddress(self.config['datastream'].data)
            dstrlabel = self.config['datastreamlabel'].data
            dstr = self.redvypr_addrconv.get_str(dstrlabel)
            self.devicedisp.setText(dstr)
            self.devicedisp.setStyleSheet("border : 1px solid {:s};".format(self.config['backgroundcolor'].data))
            self.devicedisp.setAlignment(QtCore.Qt.AlignCenter)
            self.layout.addWidget(self.devicedisp, 1, 0, 1, 2)
            self.devicedisp.show()
        else:
            self.layout.removeWidget(self.devicedisp)
            self.devicedisp.hide()

        timefmt = self.config['timeformat'].data
        if(len(timefmt)>0):
            self.timedisp.setStyleSheet("border : 1px solid {:s};".format(self.config['backgroundcolor'].data))
            self.timedisp.setText(self.get_timestr(0, format=timefmt))
            self.timedisp.setAlignment(QtCore.Qt.AlignCenter)
            self.layout.addWidget(self.timedisp, 2, 0, 1, 2)
            self.timedisp.show()
        else:
            self.layout.removeWidget(self.timedisp)
            self.timedisp.hide()

        if(len(self.config['unit'].data)>0):
            self.unitdisp.setStyleSheet("border : 1px solid {:s};".format(self.config['backgroundcolor'].data))
            self.layout.addWidget(self.unitdisp, 3, 1)
            self.unitdisp.show()
        else:
            self.layout.removeWidget(self.unitdisp)
            self.unitdisp.hide()


    def get_timestr(self,unixtime,format=None):
        """ Returns a time string
        """
        if(format == None):
            tstr = datetime.datetime.fromtimestamp(unixtime).isoformat(timespec='milliseconds')
        else:
            tstr = datetime.datetime.fromtimestamp(unixtime).strftime(format)
        return tstr
                
    def update_plot(self,data):
        """ Updates the plot based on the given data
        """
        funcname = __name__ + '.update()'
        logger.debug(funcname)
        tnow = time.time()
        #print(funcname + ': got data', data)
        dataformat = self.config['dataformat'].data
        FLAG_DATA = data in self.redvypr_addrconv
        datakey = self.redvypr_addrconv.datakey
        #print('datakey',datakey,'FLAG_DATA',FLAG_DATA)
        if(FLAG_DATA):
            # data can be a single float or a list
            newdata = data[datakey]
            newt    = data['t']
            if(type(newdata) == list):
                newdata = newdata[0]
                newt = newt[0]

            datastr = "{0:{dformat}}".format(newdata,dformat=dataformat)
            self.numdisp.setText(datastr)

            timefmt = self.config['timeformat'].data
            if(len(timefmt)>0):
                self.timedisp.setText(self.get_timestr(newt,format=timefmt))
            
            if(self.config['useprops'].data):
                # Get the propertykey 
                propkey = '?' + datakey 
                try:
                    props = data[propkey]
                except:
                    props = None
                    
            if(props is not None):
                self.unitdisp.setText(str(data[propkey]['unit'].data))
                #self.resize_font(self.unitdisp)                






    



import copy
import time
import logging
import sys
import re
import numpy as np
import redvypr
import redvypr.redvypr_address as redvypr_address
import collections
import pydantic
import typing

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('data_packets')
logger.setLevel(logging.DEBUG)

regex_symbol_start = '{'
regex_symbol_end = '}'

#device_redvypr_statdict = {'_redvypr': {}, 'datakeys': [], '_deviceinfo': {},'_keyinfo': {},'packets_received':0,'packets_published':0,'packets_droped':0}
redvypr_data_keys = ['_redvypr','_redvypr_command','_deviceinfo','_keyinfo','_metadata']

# Defintions for common metadata types
class RedvyprMetadata(pydantic.BaseModel):
    address: typing.Dict[redvypr_address.RedvyprAddress, typing.Any] = {}
class RedvyprDeviceMetadata(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    location: str = ''
    comment: str = ''

class RedvyprDatastreamMetadata(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")
    unit: str = ''
    comment: str = ''

class RedvyprMetadataGeneral(pydantic.BaseModel):
    address: typing.Dict[str, typing.Any] = {}

class Datapacket(dict):
    def __init__(self, *args, **kwargs):
        if len(args)>0:
            if isinstance(args[0],dict):
                dict.__init__(self, *args,**kwargs)

        else:
            dict.__init__(self)

        if '_redvypr' not in self.keys():
            dataself = create_datadict()
            self.update(dataself)

        self.address = redvypr_address.RedvyprAddress(self)

    def __getitem__(self, key):
        # Check if the key is a string but is an "eval" operator
        if isinstance(key,str):
            if key.startswith('[') and key.endswith(']'):
                evalstr = 'self' + key
                #data = self
                data = eval(evalstr, None)
                return data
            else:
                return super().__getitem__(key)
        # Check if the key is a RedvyprAddress
        elif isinstance(key, redvypr_address.RedvyprAddress):
            datakey = key.datakey
            if datakey.startswith('[') and datakey.endswith(']'):
                evalstr = 'self' + datakey
                #data = self
                data = eval(evalstr, None)
                return data
            else:
                return super().__getitem__(datakey)
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        # Check if the key is a string but is an "eval" operator
        if isinstance(key, str):
            if key.startswith('[') and key.endswith(']'):
                evalstr = 'self' + key + '=value'
                # data = self
                exec(evalstr, None)
            else:
                return super().__setitem__(key, value)
        # Check if the key is a RedvyprAddress
        elif isinstance(key, redvypr_address.RedvyprAddress):
            datakey = key.datakey
            if datakey.startswith('[') and datakey.endswith(']'):
                evalstr = 'self' + datakey  + '=value'
                exec(evalstr, None)
            else:
                return super().__setitem__(datakey, value)
        else:
            return super().__setitem__(key, value)

    def __expand_datakeys_recursive__(self, data, keys, level=0, parent_key='f', key_list = [], key_dict = {}, max_level=1000):
        for k in keys:
            #print('k',k)
            data_k = data[k]
            # Check if k is a list or a dict
            if isinstance(k, int):
                strformat = "[{}]".format(k)
                key_dict.append(None)
            else:
                strformat = "['{}']".format(k)
                key_dict[k] = None
            if level < max_level:
                if isinstance(data_k, list):
                    data_k_keys = range(0, len(data_k))
                    parent_key_new = parent_key + strformat
                    key_dict[k] = []
                    self.__expand_datakeys_recursive__(data_k, data_k_keys, level=level + 1, parent_key=parent_key_new, key_list=key_list, key_dict = key_dict[k], max_level=max_level)
                elif isinstance(data_k, dict):
                    data_k_keys = data_k.keys()
                    parent_key_new = parent_key + strformat
                    key_dict[k] = {}
                    self.__expand_datakeys_recursive__(data_k, data_k_keys, level=level + 1, parent_key=parent_key_new, key_list=key_list, key_dict = key_dict[k], max_level=max_level)
                elif isinstance(data_k, np.ndarray):
                    #logger.warning('Found an numpy array, this is not implemented yet')
                    # Add index type of address if necessary only
                    if level > 0:
                        expanded_key = parent_key + strformat
                        key_list.append(expanded_key)
                        key_dict[k] = (expanded_key, type(data[k]))
                    else:
                        key_list.append(k)
                        key_dict[k] = (k, type(data[k]))

                else: # This is not an iterative element anymore, lets use it
                    # Add index type of address if necessary only
                    if level > 0:
                        expanded_key = parent_key + strformat
                        key_list.append(expanded_key)
                        key_dict[k] = (expanded_key,type(data[k]))
                    else:
                        key_list.append(k)
                        key_dict[k] = (k,type(data[k]))
            else:
                # Add index type of address if necessary only
                if level > 0:
                    expanded_key = parent_key + strformat
                    key_list.append(expanded_key)
                else:
                    key_list.append(k)

    def datakeys(self, expand=False, return_type='dict'):
        """
        Returns the datakeys of the redvypr dictionary
        :param expand: True/False or level [int], if true return expanded datakeys with a level depth of 100
        :param return_type: 'dict,'list','both'
        :return:
        """
        keys = list(self.keys())
        for key_remove in redvypr_data_keys:
            try:
                keys.remove(key_remove)
            except:
                pass

        # Check if the datakey needs to be expanded
        if expand == False:
            return keys
        else:
            keys_expand = []
            keys_dict_expand = {}
            if isinstance(expand, bool):
                max_level = 100
            else:
                max_level = expand

            self.__expand_datakeys_recursive__(self, keys, level=0, parent_key='', key_list=keys_expand, key_dict=keys_dict_expand, max_level=max_level)

            if return_type=='list':
                return keys_expand
            elif return_type=='dict':
                return keys_dict_expand
            else:
                return (keys_expand, keys_dict_expand)

    def datastreams(self, expand=True):
        datakeys = self.datakeys(expand=expand,return_type='list')
        daddresses = []
        for d in datakeys:
            daddr = redvypr_address.RedvyprAddress(self.address,datakey=d)
            daddresses.append(daddr)

        return daddresses


    def legacy_datastreams_legacy(self):
        """
        Returns the datastreams of the redvypr dictionary
        """
        keys = list(self.keys())
        datastreams = []
        for key_remove in redvypr_data_keys:
            try:
                keys.remove(key_remove)
            except:
                pass

        keys.sort()
        for k in keys:
            datastreams.append(redvypr_address.RedvyprAddress(self, datakey=k))

        return datastreams

    def get_addressstr(self,addrformat='/i/k/'):
        return self.address.get_str(addrformat)




def create_datadict(data=None, datakey=None, packetid=None, tu=None, device=None, publisher=None, hostinfo=None):
    """ Creates a datadict dictionary used as internal datastructure in redvypr
    """
    if(tu == None):
        tu = time.time()

    datadict = {'_redvypr':{'t':tu}}
    datadict['_redvypr']['device'] = device
    if (packetid is None):
            datadict['_redvypr']['packetid'] = device

    datadict['_redvypr']['packetid'] = packetid
    datadict['_redvypr']['publisher'] = publisher
    if (hostinfo is not None):
        if hostinfo == 'random':
            redvypr.create_hostinfo('random')
        else:
            datadict['_redvypr']['host'] = hostinfo
    else:
        datadict['_redvypr']['host'] = redvypr.hostinfo_blank

    if(data is not None):
        if (datakey == None):
            datakey = 'data'
        datadict[datakey] = data

    return datadict


def add_metadata2datapacket(datapacket, address=None, datakey=None, metakey=None, metadata=None, metadict=None):
    """

    Args:add_metad
        datapacket:
        datakey:
        metakey:
        metadata:

    Returns:

    """
    if True:
        #print('Datapacket',datapacket)
        if address is not None:
            raddress = redvypr_address.RedvyprAddress(address)
        if datakey is not None:
            try: # Try first to create a RedvyprAddress from the datapacket itself
                raddress = redvypr_address.RedvyprAddress(datapacket, datakey=datakey)
            except:
                logger.info('Could not create address',exc_info=True)
                raddress = redvypr_address.RedvyprAddress(datakey=datakey)

        address_str = raddress.address_str
        try:
            datapacket['_metadata']
        except:
            datapacket['_metadata'] = {}

        try:
            datapacket['_metadata'][address_str]
        except:
            datapacket['_metadata'][address_str] = {}

        if (metadata is not None):
            datapacket['_metadata'][address_str][metakey] = metadata

        # If a dictionary with metakeys is given
        if (metadict is not None):
            datapacket['_metadata'][address_str].update(metadict)

    #print('Metadata datapacket',datapacket)
    #print('---done----')
    return datapacket



def commandpacket(command='stop',device_uuid='',thread_uuid='',packetid=None,devicename=None,publisher=None,host=None,comdata=None,devicemodulename=None):
    """

    Args:
        command: 'stop'
        device: The device the command was sent from
        device_uuid:
        thread_uuid:

    Returns:
         compacket: A redvypr dictionary with the command
    """
    compacket = create_datadict({'command':command}, datakey='_redvypr_command') # The command
    if packetid is not None:
        compacket['_redvypr']['packetid'] = packetid  # The device the command was sent from
    if devicename is not None:
        compacket['_redvypr']['device'] = devicename  # The device the command was sent from
    if publisher is not None:
        compacket['_redvypr']['publisher'] = publisher
    if host is not None:
        compacket['_redvypr']['host'] = host
    if devicemodulename is not None:
        compacket['_redvypr']['devicemodulename'] = devicemodulename  # The device the command was sent from
    compacket['_redvypr_command']['device_uuid'] = device_uuid  # The uuid of the device the command is for
    compacket['_redvypr_command']['thread_uuid'] = thread_uuid  # The uuid of the thread of device the command is for
    compacket['_redvypr_command']['data'] = comdata

    return compacket


def deviceinfopacket(deviceadress,statusdict):
    """
    deviceinfopacket for a device thread
    Returns:
         stauspacket: A redvypr statuspacket dictionary
    """
    comdata = {}
    comdata['deviceaddr'] = deviceadress
    comdata['devicestatus'] = statusdict  # This is the status of the device
    datapacket = commandpacket(command='device_status', device_uuid='', thread_uuid='', devicename=None,
                                            host=None, comdata=comdata)
    return datapacket


def statuspacket(deviceadress,statusdict):
    """
    statuspacket for a device thread
    Returns:
         statuspacket: A redvypr statuspacket dictionary
    """
    comdata = {}
    comdata['deviceaddr'] = deviceadress
    comdata['devicestatus'] = statusdict  # This is the status of the device
    datapacket = commandpacket(command='device_status', device_uuid='', thread_uuid='', devicename=None,
                                            host=None, comdata=comdata)
    return datapacket

def check_for_command(datapacket=None,uuid=None,thread_uuid=None,add_data=False):
    """

    Args:
        datapacket:
        uuid: if set, compare uuid of the command with given uuid, return command only of uuid match
        add_data: adds the command dictionary

    Returns:
        command: content of the field 'redvypr_device_command', typically a string
    """
    flags = {}
    if '_redvypr_command' not in datapacket.keys():
        if(add_data):
            return [None,None]
        else:
            return None
    else:
        FLAG_COM1 = 'command' in datapacket['_redvypr_command'].keys()
        FLAG_COM2 = 'device_uuid' in datapacket['_redvypr_command'].keys()
        FLAG_COM3 = 'thread_uuid' in datapacket['_redvypr_command'].keys()
        command = None
        flags['com'] = FLAG_COM1
        if FLAG_COM1:

            if(uuid is not None):
                FLAG_UUID = datapacket['_redvypr_command']['device_uuid'] == uuid
            else:
                FLAG_UUID = True

            if (thread_uuid is not None):
                FLAG_TUUID = datapacket['_redvypr_command']['thread_uuid'] == thread_uuid
            else:
                FLAG_TUUID = True

            flags['device_uuid'] = FLAG_UUID
            flags['thread_uuid'] = FLAG_TUUID
            command = datapacket['_redvypr_command']['command']

        if(add_data):
            return [command,{'command_data':datapacket['_redvypr_command'],'flags':flags}]
        else:
            return command


#__rdvpraddr__ = redvypr_address('tmp')
#addresstypes  = __rdvpraddr__.get_strtypes() # A list of all addresstypes




"""
Sensor definitions
"""
import typing
import datetime
import dateutil
import pydantic
import struct
import re
import sys
import numpy
import copy
import logging
from redvypr.data_packets import create_datadict as redvypr_create_datadict, add_metadata2datapacket, Datapacket
from redvypr.redvypr_address import RedvyprAddress
from redvypr.devices.sensors.calibration.calibration_models import calibration_HF, calibration_NTC, calibration_const, \
    calibration_poly


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('sensor_definitions')
logger.setLevel(logging.DEBUG)

def decode_utf8(byte_string):
    return byte_string.decode('utf-8')


def array(byte_string):
    return list(numpy.fromstring(byte_string,sep=','))



class Sensor(pydantic.BaseModel):
    name: str = pydantic.Field(default='sensor')
    description: str = pydantic.Field(default='Sensor')
    example_data: typing.Union[None,bytes, str] = pydantic.Field(default=None)
    sensortype: typing.Literal['sensor'] = pydantic.Field(default='sensor')
    datastream: RedvyprAddress = pydantic.Field(default=RedvyprAddress('*'))
    autofindcalibration: bool = pydantic.Field(default=True, description='Tries to find automatically calibrations for the sensor')
    calibrations: typing.Dict[str, typing.Annotated[typing.Union[calibration_const, calibration_poly], pydantic.Field(
        discriminator='calibration_type')]] = pydantic.Field(default={})

    def add_calibration_for_datapacket(self, packetaddress=RedvyprAddress('/i:*'), addrformat='/i',calibration=None):
        """
        Adds a calibration to the packetaddress. During process data it is checked if the provided datapacket fits with the packetaddress of the calibration and if so all calibrations found are applied.
        :param packetaddress:
        :param calibration:
        :return:
        """
        if calibration is not None:
            key = packetaddress.get_str(addrformat)
            try:
                self.calibrations[key]
            except:
                self.calibrations[key] = []

            self.calibrations[key].append(calibration)

    def add_all_calibrations(self, calibrations):
        """
        Adding a list of all calibrations to sensor. The calibrations are used for the automatic search for the correct calibrations for the processed sensor data
        :param calibrations:
        :return:
        """
        self.__all_calibrations = calibrations


    def find_calibration_for_datapacket(self, rdata: Datapacket):
        """
        Tries to find the right calibration for the datapacket
        :param rdata:
        :return:
        """
        funcname = __name__ + '.find_calibration_for_datapacket():'
        logger.debug(funcname)
        # TODO: Find a smart way to check if an autocalibration has been already done
        try:
            self.__datapackets_checked_for_calibrations
        except:
            self.__datapackets_checked_for_calibrations = []

        self.__datapackets_checked_for_calibrations.append(rdata.address)
        for calibration in self.__all_calibrations:
            logger.debug(funcname + 'Checking calibration {}'.format(calibration))
            sn = calibration.sn
            sensor_model = calibration.sensor_model
            caldate = calibration.date
            parameter_calibrated = calibration.parameter
            # 1: Check if calibration.parameter is existing in the datapacket
            # 2: Compare serial number with packetid (not done yet)
            # 3: save date (not done yet)
            try:
                #print('parameter calibrated', parameter_calibrated)
                caldata_raw = rdata[parameter_calibrated]
                #print('Got caldata raw', caldata_raw)
                flag_parameter = True
            except:
                logger.debug(funcname + 'Could not get data', exc_info=True)
                flag_parameter = False

            if flag_parameter:
                logger.debug(funcname+ 'Adding calibration')
                calibration_apply = calibration.model_copy()
                calibration_apply.address_apply = calibration_apply.parameter
                #calibration_apply.datakey_result = 'test1'
                self.add_calibration_for_datapacket(rdata.address,calibration=calibration_apply)


    def datapacket_process(self, data, check_own_address=True):
        """
        Processes a redvypr datapacket. This is mainly applying calibrations
        :param data:
        :return:
        """
        funcname = __name__ + '.datapacket_process():'

        if data in self.datastream or (check_own_address==False):
        #if True: # self.datastream does not work for binary sensors
            rdata = Datapacket(data)
            rdata_addressstr = rdata.get_addressstr('/i')
            # Check if autofindcalibration shall be done
            if self.autofindcalibration:
                #print('Autocalibration')
                try:
                    self.__datapackets_checked_for_calibrations
                except:
                    self.__datapackets_checked_for_calibrations = []

                if rdata.address in self.__datapackets_checked_for_calibrations:
                    #print('Datapacket was already checked, doing nothing')
                    pass
                else:
                    found_calibration = False
                    for datapacket_calkey in self.calibrations.keys():
                        datapacket_calkey_address = RedvyprAddress(datapacket_calkey)
                        if rdata in datapacket_calkey_address:
                            found_calibration = True
                            break

                    # Did not find a calibration, but it is wished to search for one
                    if found_calibration == False:
                        self.find_calibration_for_datapacket(rdata)
                    else:
                        pass
                        #print('Found calibration already')

            # Check if there is a calibration to be found for the datapacket
            #print('Calibrations',self.calibrations)
            if len(self.calibrations.keys()) == 0:
                #print('Returning data')
                return data
            else:
                for datapacket_calkey in self.calibrations.keys():
                    datapacket_calkey_address = RedvyprAddress(datapacket_calkey)
                    if rdata in datapacket_calkey_address:
                        calibrations_for_packet = self.calibrations[rdata_addressstr]
                        #print('Calibrations for packet',calibrations_for_packet)
                        # Loop over all calibrations for the datapacket
                        for calibration in calibrations_for_packet:
                            #print('Processing calibration', calibration)
                            caldata_raw = rdata[calibration.address_apply]
                            #print(caldata_raw)
                            # And now the most important thing, applying the calibration
                            caldata_cal = calibration.raw2data(caldata_raw)
                            #print(caldata_raw,caldata_cal)
                            # Copy the data to the original location or make a copy
                            if calibration.datakey_result is None:
                                #print('Applying to data')
                                rdata[calibration.address_apply] = caldata_cal
                            else: # If the base datakey shall be changed, copy the data first
                                if calibration.address_apply.parsed_addrstr_expand['datakeyeval']:
                                    datakey_orig = calibration.address_apply.parsed_addrstr_expand['datakeyentries'][0]
                                    datakey_new = calibration.datakey_result.format(datakey=datakey_orig)
                                    datakey_new_eval = calibration.address_apply.datakey.replace(datakey_orig,"{}".format(datakey_new))
                                    address_result = RedvyprAddress(calibration.address_apply, datakey=datakey_new_eval)
                                else:
                                    datakey_orig = calibration.address_apply.datakey
                                    # Create the new RedvyprAddress
                                    datakey_new = calibration.datakey_result.format(datakey=datakey_orig)
                                    address_result = RedvyprAddress(calibration.address_apply, datakey=datakey_new)

                                # And finally copy the data into the new datakey
                                try:
                                    rdata[datakey_new]
                                except:
                                    data_orig = data[datakey_orig]
                                    rdata[datakey_new] = copy.deepcopy(data_orig)

                                #print('Data orig',data_orig)
                                #print('Hallo',calibration.address_apply.parsed_addrstr_expand['datakeyentries'])
                                rdata[address_result] = caldata_cal

                        #print('data done', rdata)
                        return dict(rdata)


class BinarySensor(Sensor):
    """
    A binary sensor gets binary data that need to be converted first into a dictionary with the datakeys
    """
    name: str = pydantic.Field(default='binsensor')
    sensortype: typing.Literal['binsensor'] = pydantic.Field(default='binsensor')
    regex_split: bytes = pydantic.Field(default=b'', description='Regex expression to split a binary string')
    binary_format: typing.Dict[str, str] = pydantic.Field(default={}, description='https://docs.python.org/3/library/struct.html, for example for 16bit signed data {"adc_data":"<h"}')
    str_format: typing.Dict[str, str] = pydantic.Field(default={})
    packetid_format: typing.Optional[str] = pydantic.Field(default=None,description='Format of the packetid of the datapacket')
    #packetid_format: typing.Union[float,str, None] = pydantic.Field(default=3.0,
    #                                                       description='Format of the packetid of the datapacket')
    datakey_metadata: typing.Dict[str, typing.Dict] = pydantic.Field(default={})
    calibrations_raw: typing.Dict[str, typing.Annotated[typing.Union[calibration_const, calibration_poly], pydantic.Field(
        discriminator='calibration_type')]] = pydantic.Field(default={})

    calibration_python_str: typing.Optional[dict] = pydantic.Field(default=None, description='A python str that is evaluated and processes the raw data')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__rdatastream__ = RedvyprAddress(self.datastream)
        self._str_functions = {}
        self._str_functions_invalid_data = {}
        self._flag_binary_keys = len(self.binary_format.keys()) > 0
        self._flag_str_format_keys = len(self.str_format.keys()) > 0
        # Add functions for datatypes
        for key in self.str_format:
            vartype = self.str_format[key]
            #print('key', key, 'vartype', vartype)
            if vartype.lower() == 'float':
                self._str_functions[key] = float
                self._str_functions_invalid_data[key] = None
            elif vartype.lower() == 'int':
                self._str_functions[key] = int
                self._str_functions_invalid_data[key] = None
            elif vartype.lower() == 'str':
                self._str_functions[key] = decode_utf8
                self._str_functions_invalid_data[key] = None
            elif vartype.lower() == 'array':
                self._str_functions[key] = array
                self._str_functions_invalid_data[key] = None

    def create_metadata_datapacket(self):
        """
        Creates a datapacket with the metadata information
        :return:
        """
        data_packet = redvypr_create_datadict(device=self.name)

        flag_metadata = False
        for key_input in self.datakey_metadata.keys():
            flag_metadata = True
            metadata = self.datakey_metadata[key_input]
            metadata_address = RedvyprAddress(devicename=self.name,datakey=key_input)
            #data_packet = add_metadata2datapacket(data_packet, key_input, metadict=metadata)
            data_packet = add_metadata2datapacket(data_packet, address=metadata_address, metadict=metadata)

        for key_input in self.calibrations_raw.keys():
            flag_metadata = True

            calibration = self.calibrations_raw[key_input]
            unit = calibration.unit
            unit_input = calibration.unit_input
            key_result = calibration.parameter_result
            if (key_input is not None) and (unit_input is not None):
                metadata_address = RedvyprAddress(devicename=self.name, datakey=key_input)
                data_packet = add_metadata2datapacket(data_packet, address=metadata_address, metadata=unit_input)
                #data_packet = add_metadata2datapacket(data_packet, key_input, metadata=unit_input)
            if (key_result is not None) and (unit is not None):
                metadata_address = RedvyprAddress(devicename=self.name, datakey=key_result)
                data_packet = add_metadata2datapacket(data_packet, address=metadata_address, metadata=unit)


        if flag_metadata:
            return data_packet
        else:
            return None
    def datapacket_process(self, data):
        """
        Processes a redvypr datapacket.
        :param data:
        :return:
        """
        #print('Hallo data', data)
        if data in self.__rdatastream__:
            #print('Processing data',data)
            binary_data = self.__rdatastream__.get_data(data)
            #print('Binary data', binary_data)
            data_packets = self.binary_process(binary_data, datapacket_orig=data)
            #print('data packets',data_packets)
            for data_packet in data_packets:
                if 't' not in data_packet.keys():
                    data_packet['t'] = data['t']

            # Check if calibrations exist, if yes, do the calibration procedure
            if (len(self.calibrations.keys()) > 0) or self.autofindcalibration:
                #print('Calibrations',self.calibrations)
                #print('Autofindcalibration',self.autofindcalibration)
                data_packets_calibrated = []
                for data_packet in data_packets:
                    #print('Found calibration for parameter')
                    data_packet = super().datapacket_process(data_packet, check_own_address=False)
                    data_packets_calibrated.append(data_packet)

                #print('Autocalibration',data_packets_calibrated)
                return data_packets_calibrated
            else: # no calibration, just return the data packets
                return data_packets

    def binary_process(self, binary_stream, datapacket_orig):
        """

        """
        rematches = self.binary_split(binary_stream)
        data_packets = []
        #print('Rematches',rematches)
        # Loop over all split raw string packages
        for rematch in rematches:
            data_packet = redvypr_create_datadict(device=self.name)
            flag_data = False
            #print('Processing match', rematch)
            #print('Variables found', rematch.groupdict())
            redict = rematch.groupdict()
            if self._flag_binary_keys:
                for keyname in redict:
                    if keyname in self.binary_format.keys():
                        binary_format = self.binary_format[keyname]
                        #print('Found binary key with format', keyname, binary_format)
                        # convert the data
                        data = struct.unpack(binary_format, redict[keyname])
                        if len(data) == 1:
                            data = data[0]
                        data_packet[keyname] = data
                        flag_data = True


            if self._flag_str_format_keys:
                #print('Str format')
                for keyname in redict:
                    if keyname in self.str_format.keys():
                        #print('Found str key', keyname)
                        # get the right function
                        convfunction = self._str_functions[keyname]
                        # convert the data, if this fails, take invalid data value
                        try:
                            data = convfunction(redict[keyname])
                        except:
                            #print('Data',redict[keyname])
                            logger.debug('Could not decode data for key {}'.format(keyname),exc_info=True)
                            data = self._str_functions_invalid_data[keyname]

                        data_packet[keyname] = data
                        flag_data = True
                        #print('Converted data to', data, flag_data)

            if self.calibration_python_str is not None:
                #print('Found a python calibration eval str, applying')
                for keyname_eval in self.calibration_python_str:
                    evalcommand = self.calibration_python_str[keyname_eval]
                    evalcommand = evalcommand.split('\n')[0]
                    evalstr_full = 'data_packet[keyname_eval]=' + evalcommand
                    #print('Evalstr full',evalstr_full)
                    try:
                        exec(evalstr_full)
                    except:
                        logger.info('Could not evaluate command',exc_info=True)
                    #print('Data packet',data_packet)


            # Check for calibrations
            datakeys = list(data_packet.keys())
            for keyname in datakeys:
                # Check if there is a calibration
                if keyname in self.calibrations_raw.keys():
                    data = data_packet[keyname]
                    #print('Found a calibration to convert raw data for {}'.format(keyname))
                    calibration = self.calibrations_raw[keyname]
                    try:
                        keyname_cal = calibration.datakey_result
                        if keyname_cal is None:
                            keyname_cal = keyname
                    except:
                        keyname_cal = keyname + '_cal'

                    data_cal = calibration.raw2data(data)
                    #print('Data cal',data_cal)
                    data_packet[keyname_cal] = data_cal

            # Check for packetid
            if self.packetid_format is not None:
                # Define variables that can be used for the packetid, see RMC as an example
                datapacket_orig_dict = {}
                datapacket_orig_dict['packetid_rawdata'] = datapacket_orig['_redvypr']['packetid']
                datapacket_orig_dict['publisher_rawdata'] = datapacket_orig['_redvypr']['publisher']
                datapacket_orig_dict['device_rawdata'] = datapacket_orig['_redvypr']['device']
                try:
                    packetidstr = self.packetid_format.format(**data_packet, **datapacket_orig_dict)
                    data_packet['_redvypr']['packetid'] = packetidstr
                except:
                    logger.warning('Could not create an packetidstr:',exc_info=True)
            else:
                packetidstr = self.name
                data_packet['_redvypr']['packetid'] = packetidstr

            #print('Test flag data',flag_data)
            if flag_data:
                data_packets.append(data_packet)

        #print('Data packets',data_packets)
        return data_packets

    def binary_split(self, binary_stream):
        """
        Splits the data into pieces
        :param binary_stream:
        :param sensors:
        :return:
        """

        regex = self.regex_split
        matches = []
        #print('Regex', regex, binary_stream)
        # rematch = re.search(regex, binary_stream)
        rematchiter = re.finditer(regex, binary_stream)
        rematch = [r for r in rematchiter]
        #print('rematch 1', rematch)
        return rematch



# S4L (sam4log)
s4l_split = b'B\x00(?P<counter32>[\x00-\xFF]{4})(?P<adc16>[\x00-\xFF]{2})\n'
s4l_binary_format = {'counter32': '<L','adc16':'<h'}
Vref = 3.0 # Reference voltage in V
coeff_fac = Vref/2**15
coeff_fac_counter = 1/1024
calibration_adc16 = calibration_const(parameter_result='adc(V)',coeff=coeff_fac,
                                      unit='V',unit_input='counts')
calibration_counter32 = calibration_const(parameter_result='counter(s)',coeff=coeff_fac_counter,
                                      unit='s',unit_input='counts')
calibrations_raw = {'adc16':calibration_adc16,'counter32':calibration_counter32}
S4LB = BinarySensor(name='S4LB', regex_split=s4l_split, binary_format=s4l_binary_format,
                    datastream=RedvyprAddress('/k:data'),
                    calibrations_raw=calibrations_raw)


# NMEA RMC
#https://de.wikipedia.org/wiki/NMEA_0183#Recommended_Minimum_Sentence_C_(RMC)
nmea_rmc_description = 'NMEA Recommended Minimum Sentence\nhttps://de.wikipedia.org/wiki/NMEA_0183#Recommended_Minimum_Sentence_C_(RMC)'
#nmea_rmc_split = b'\$[A-Z]+RMC,(?P<time>[0-9.]*),(?P<status>[A-Z]+),(?P<latdeg>[0-9]{2})(?P<latmin>[0-9.]+),(?P<NS>[NS]+),(?P<londeg>[0-9]{3})(?P<lonmin>[0-9.]+),(?P<EW>[EW]+),[0-9.]*,[0-9.]*,(?P<date>[0-9.]*),.*\n'
nmea_rmc_split = b'\$(?P<devid>[A-Z]+)RMC,(?P<time>[0-9.]+),(?P<status>[A-Z]+),(?P<latstr>[0-9.]*),(?P<NS>[NS]*),(?P<lonstr>[0-9.]*),(?P<EW>[EW]*),(?P<speed>[0-9.]*),(?P<course>[0-9.]*),(?P<date>[0-9.]+),(?P<magdev>[0-9.]*),(?P<magdevdir>[EW]*),(?P<crc>.*)\n'
nmea_rmc_str_format = {'devid':'str','time':'str','date':'str','latstr':'str','lonstr':'str','NS':'str','EW':'str','speed':'float','course':'float'}
nmea_datakey_metadata = {'time':{'unit':'HHMMSS','description':'GNSS in UTC'},'lat':{'unit':'degN'},'lon':{'unit':'degE'}}
nmea_rmc_test1 = b'$GPRMC,162614.22,A,5230.5900,N,01322.3900,E,10.0,90.0,131006,1.2,E,A*13\r\n'
#nmea_rmc_test2 = b'$GPRMC,090413.788,V,,,,,,,310724,,,N*46\r\n'
nmea_rmc_test2 = b'$GPRMC,090413,V,,,,,,,310724,,,N*46\r\n'
nmea_rmc_packetid_format = 'RMC_{packetid_rawdata}_{publisher_rawdata}_{devid}'
latevalstr = '(float(data_packet["latstr"][0:2])+float(data_packet["latstr"][2:])/60)*(float(data_packet["NS"]=="N")-float(data_packet["NS"]=="S"))'
lonevalstr = '(float(data_packet["lonstr"][0:3])+float(data_packet["lonstr"][3:])/60)*(float(data_packet["EW"]=="W")-float(data_packet["EW"]=="E"))'
timeevalstr = 'dateutil.parser.parse(data_packet["date"] + " " + data_packet["time"] + "UTC",yearfirst=False).timestamp()'
nmea_calibration_python_str = {'lat':latevalstr,'lon':lonevalstr,'t':timeevalstr}
NMEARMC = BinarySensor(name='NMEA0183_RMC', regex_split=nmea_rmc_split,
                       str_format=nmea_rmc_str_format,
                       description=nmea_rmc_description,
                       example_data=nmea_rmc_test1,
                       datastream=RedvyprAddress('/k:data'),
                       datakey_metadata=nmea_datakey_metadata,
                       packetid_format=nmea_rmc_packetid_format,
                       calibration_python_str=nmea_calibration_python_str)


# TAR (temperature array)
tar_b2_test1 = b'$FC0FE7FFFE155D8C,TAR,B2,36533.125000,83117,3498.870,3499.174,3529.739,3490.359,3462.923,3467.226,3480.077,3443.092,3523.642,3525.567,3509.492,3561.330,3565.615,3486.693,3588.670,3539.169,3575.104,3523.946,3496.343,3480.160,3531.045,3501.624,3497.010,3557.235,3479.952,3458.297,3523.052,3487.223,3571.087,3525.740,3580.928,3534.818\n'
#tar_b2_split = b'\$(?P<MAC>[A-F,0-9]+),TAR,B2,(?P<counter>[0-9.]+),(?P<np>[0.9]+),(?P<TAR>[0-9.]+,*)\n'
tar_b2_split = b'\$(?P<MAC>.+),TAR,B2,(?P<counter>[0-9.]+),(?P<np>[0-9]+),(?P<TAR>.*)\n'
tar_b2_str_format = {'MAC':'str','counter':'float','np':'int','TAR':'array'}
tar_b2_datakey_metadata = {'MAC':{'unit':'MAC64','description':'MAC of the sensor'},'np':{'unit':'counter'},'TAR':{'unit':'Ohm'}}
tar_b2_packetid_format = 'TAR_B2_{MAC}'
tar_b2_description = 'Temperature array NMEA like text format'
tar_b2 = BinarySensor(name='tar_b2', regex_split=tar_b2_split,
                       str_format=tar_b2_str_format,
                       description=tar_b2_description,
                       example_data=tar_b2_test1,
                       datakey_metadata=tar_b2_datakey_metadata,
                       packetid_format=tar_b2_packetid_format,
                       datastream=RedvyprAddress('/k:data'))


# HF (Heatflow)
HF_test1 = b'$FC0FE7FFFE1567E3,HF,00000431.3125,108,-0.000072,2774.364,2782.398,2766.746\n'
HF_test2 = b'$FC0FE7FFFE1567E3,HF,00054411.3125,13603,0.000037,2780.217,2786.642,2774.316\n'
HF_split = b'\$(?P<MAC>.+),HF,(?P<counter>[0-9.]+),(?P<np>[0-9]+),(?P<HF_V>[-,\+]*[0-9.]+),(?P<NTC_R>.*)\n'
HF_str_format = {'MAC':'str','counter':'float','np':'int','HF_V':'float','NTC_R':'array'}
HF_datakey_metadata = {'MAC':{'unit':'MAC64','description':'MAC of the sensor'},'np':{'unit':'counter'},'HF_V':{'unit':'Volt'},'NTC_R':{'unit':'Ohm'}}
HF_packetid_format = 'HF_{MAC}'
HF_description = 'Heatflow sensor raw data (units are Volt and Ohm)'
HF = BinarySensor(name='HF', regex_split=HF_split,
                       str_format=HF_str_format,
                       description=HF_description,
                       example_data=HF_test1,
                       datakey_metadata=HF_datakey_metadata,
                       packetid_format=HF_packetid_format,
                       datastream=RedvyprAddress('/k:data'))


# HFS (Heatflow)
HFS_test1 = b'$FC0FE7FFFE153BDC,HFS,00182595.0000,45649,2.151728,29.609,29.561,29.631\n'
HFS_test2 = b'$FC0FE7FFFE153BDC,HFS,00182635.0000,45659,2.152100,29.609,29.560,29.631\n'
HFS_split = b'\$(?P<MAC>.+),HFS,(?P<counter>[0-9.]+),(?P<np>[0-9]+),(?P<HF_Wm2>[-,\+]*[0-9.]+),(?P<T_degC>.*)\n'
HFS_str_format = {'MAC':'str','counter':'float','np':'int','HF_Wm2':'float','T_degC':'array'}
HFS_datakey_metadata = {'MAC':{'unit':'MAC64','description':'MAC of the sensor'},'np':{'unit':'counter'},'HF_Wm2':{'unit':'W m-2'},'T_degC':{'unit':'degC'}}
HFS_packetid_format = 'HFS_{MAC}'
HFS_description = 'Heatflow sensor data in SI units'
HFS = BinarySensor(name='HFS', regex_split=HFS_split,
                       str_format=HFS_str_format,
                       description=HFS_description,
                       example_data=HFS_test1,
                       datakey_metadata=HFS_datakey_metadata,
                       packetid_format=HFS_packetid_format,
                       datastream=RedvyprAddress('/k:data'))


predefined_sensors = []
predefined_sensors.append(HF)
predefined_sensors.append(HFS)
predefined_sensors.append(S4LB)
predefined_sensors.append(NMEARMC)
predefined_sensors.append(tar_b2)



import redvypr
import time

print('This script demonstrates the conceptual creation, processing and filter steps of a redvypr datapacket')

print('Creating hostinfo')
hostinfo = redvypr.create_hostinfo(hostname='someredvypr')
hostinfo2 = redvypr.create_hostinfo(hostname='otherredvypr')

# Create bogus devicenames, devicemodulename and numpackets
devicename = 'somedevice'
packetid = 'someid'
devicemodulename = 'somedevicemodulename'
numpacket = 0
tread = time.time() # The time the packet was received from the queue (in redvypr.distribute_data)
# Create a dictionary with some data
data = {'x':10,'y':20,'z':[1,2,3,4],'u':{'a':[5,6,7],'b':'Hello'}}
print('Data sent by the device thread:')
print(data)
print('Treating the datapacket, i.e. adding hostinfo, devicename ...')
redvypr.redvypr_packet_statistic.treat_datadict(data, devicename, hostinfo, numpacket, tread,devicemodulename)
print('Data after being received by the redvypr main thread and garnished with additional information:')
print(data)

print('And now send the packet through a second redvypr instance')
devicename2 = 'otherdevice'
packetid2 = 'otherid'
devicemodulename2 = 'otherdevicemodulename'
numpacket2 = 1000
tread2 = time.time()
redvypr.redvypr_packet_statistic.treat_datadict(data, devicename2, hostinfo2, numpacket2, tread2, devicemodulename2)
print('Data after received by the second redvypr main thread and garnished with additional information:')
print(data)


print('And now get the datakeys')
rdata = redvypr.data_packets.Datapacket(data)
(datakeys,datakeys_dict) = rdata.datakeys(expand=True,return_type='both')
print('Datakeys',datakeys)
print('Datakeys',datakeys_dict)


import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import yaml
import copy
import pyqtgraph
import pyqtgraph.dockarea
import qtawesome
import pydantic
import typing
import redvypr.data_packets
from redvypr.device import RedvyprDevice
from redvypr.gui import iconnames
#from redvypr.gui import configWidget
#from redvypr.devices.plot.plot_widgets import redvypr_numdisp_widget, redvypr_graph_widget, config_template_numdisp, config_template_graph
from redvypr.devices.plot.XYplotWidget import XYPlotWidget
import redvypr.files as files
from redvypr.device import device_start_standard
from redvypr.data_packets import check_for_command
#from redvypr.redvypr_packet_statistic import do_data_statistics, create_data_statistic_dict
#from redvypr.configdata import configdata, getdata

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('plot')
logger.setLevel(logging.DEBUG)

redvypr_devicemodule = True

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = True
    description: str = 'Allows to send data manually'
    gui_icon: str = 'ph.chart-line-fill'


class PlotConfig(pydantic.BaseModel):
    plottype: typing.Literal['XY-Plot'] = pydantic.Field(default='XY-Plot')
    docklabel: str = pydantic.Field(default='Dock')
    location: typing.Literal['left','right','top','bottom'] = pydantic.Field(default='left')


class DeviceCustomConfig(pydantic.BaseModel,extra='allow'):
    #plots: list = pydantic.Field(default=[])
    plots: typing.Optional[typing.Dict[str,PlotConfig]] = pydantic.Field(default={}, editable=True)
    dockstate: typing.Optional[dict] = pydantic.Field(default=None)


# Use the standard start function as the start function
start = device_start_standard

class Device(RedvyprDevice):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

    def get_config(self):
        config = super().get_config()
        state = self.redvyprdevicewidget.dockarea.saveState()
        config.custom_config.dockstate = state
        return config


class PlotDock(pyqtgraph.dockarea.Dock):
    labelClicked = QtCore.pyqtSignal()  #
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs, closable=True)
        self.label.sigClicked.connect(self.__labelClicked)

    def __labelClicked(self):
        self.labelClicked.emit()

    def mousePressEvent(self, ev):
        super().mousePressEvent(ev)

    def mouseDoubleClickEvent(self, ev):
        super().mouseDoubleClickEvent(ev)


class RedvyprDeviceWidget(redvypr.widgets.standard_device_widgets.RedvyprDeviceWidget_startonly):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        layout = QtWidgets.QGridLayout(self)
        self.dockarea = pyqtgraph.dockarea.DockArea()
        self.plot_widgets = {}
        self.numdocks = 1
        if len(self.device.custom_config.plots.keys()) == 0:
            # This adds a plot, if not existing already
            xyplotconfig = PlotConfig(plottype='XY-Plot',docklabel='Dock 1')
            self.add_plot_to_config(xyplotconfig)

        self.add_plots_from_config()
        menu = QtWidgets.QMenu()
        menu_XYplot = QtWidgets.QMenu('XY-Plot',self)
        menu.addMenu(menu_XYplot)
        addAction = QtWidgets.QWidgetAction(self)
        # Create add XYPlot-Menu
        addMenuWidget = QtWidgets.QWidget()
        addMenuWidget_layout = QtWidgets.QVBoxLayout(addMenuWidget)
        comboXYplotlocation = QtWidgets.QComboBox()
        comboXYplotlocation.addItem('Left')
        comboXYplotlocation.addItem('Right')
        comboXYplotlocation.addItem('Top')
        comboXYplotlocation.addItem('Bottom')
        addButton = QtWidgets.QPushButton('Add')
        addButton.clicked.connect(self.__add_xy_clicked)
        dockname_str = "Dock {}".format(self.numdocks)
        dockName = QtWidgets.QLineEdit(dockname_str)
        addButton.__comboXYplotlocation = comboXYplotlocation
        addButton.__dockName = dockName
        addButton.__addwidget_bare = XYPlotWidget
        addButton.__plottype = 'XY-Plot'
        addMenuWidget_layout.addWidget(dockName)
        addMenuWidget_layout.addWidget(comboXYplotlocation)
        addMenuWidget_layout.addWidget(addButton)
        addAction.setDefaultWidget(addMenuWidget)
        menu_XYplot.addAction(addAction)

        #icon = qtawesome.icon(iconnames['settings'])
        # self.settings_button = QtWidgets.QPushButton('Settings')
        # self.settings_button.setIcon(icon)
        self.add_button = QtWidgets.QPushButton('Add')
        self.add_button.setMenu(menu)


        #self.settings_button.clicked.connect(self.settings_clicked)
        self.layout.removeWidget(self.buttons_widget)
        self.layout.removeWidget(self.killbutton)
        self.killbutton.hide()

        self.layout.addWidget(self.dockarea,0,0)
        self.layout.addWidget(self.buttons_widget, 1, 0)
        self.layout_buttons.addWidget(self.add_button, 0, 4)
        #self.layout_buttons.addWidget(self.settings_button, 0, 6)

    def add_plot_to_config(self, plotconfig):
        funcname = __name__ + '.add_plot_to_config():'
        docklabel = plotconfig.docklabel
        if docklabel in self.device.custom_config.plots:
            logger.warning(funcname + 'Could not add plot {}'.format(docklabel))
        else:
            self.device.custom_config.plots[docklabel] = plotconfig
    def add_plot_from_config(self, plotconfig):
        funcname = __name__ + '.add_plot_from_config():'
        docklabel = plotconfig.docklabel
        if docklabel in self.plot_widgets:
            logger.warning(funcname + 'Plot {} exists already'.format(docklabel))
        else:
            logger.debug(funcname + 'Adding plot {}'.format(plotconfig))
            # Test the dockarea
            dock = PlotDock(plotconfig.docklabel, size=(1, 1))  # give this dock the minimum possible size
            if plotconfig.plottype == 'XY-Plot':
                plotwidget = XYPlotWidget(redvypr_device=self.device)
            else:
                raise 'Unknown plottype'

            self.plot_widgets[docklabel] = plotwidget
            dock.addWidget(plotwidget)
            self.dockarea.addDock(dock, plotconfig.location)
            self.numdocks += 1

    def add_plots_from_config(self):
        funcname = __name__ + '.add_plots_from_config():'
        for docklabel in self.device.custom_config.plots:
            d = self.device.custom_config.plots[docklabel]
            self.add_plot_from_config(d)

        dockstate = self.device.custom_config.dockstate
        if dockstate is not None:
            logger.debug(funcname + 'Found dockstate, will apply')
            self.dockarea.restoreState(dockstate)
            # Remove dockstate
            self.device.custom_config.dockstate = None

    def __add_xy_clicked(self):
        button = self.sender()
        position = button.__comboXYplotlocation.currentText().lower()
        dockname_str = button.__dockName.text()
        plottype_str = button.__plottype
        plotconfig = PlotConfig(plottype=plottype_str, docklabel=dockname_str,location=position)
        logger.debug('Adding plot {}'.format(plotconfig))
        self.add_plot_to_config(plotconfig)
        self.add_plots_from_config()
        dockname_str = "Dock {}".format(self.numdocks)
        button.__dockName.setText(dockname_str)

    def __dockclosed(self):
        dock = self.sender()
        for w in dock.widgets:
            for k in self.plot_widgets:
                if w == self.plot_widgets[k]:
                    wdict = self.plot_widgets.pop(k)
                    w.close()
                    break

    def update_data(self, data):
        for w in self.plot_widgets:
            w.update_plot(data)






import copy
import datetime
import os.path
import zoneinfo
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import pyqtgraph
import yaml
import uuid
import matplotlib
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import random
import pydantic
import typing
from redvypr.redvypr_address import RedvyprAddress
from redvypr.data_packets import check_for_command
from redvypr.device import RedvyprDevice, RedvyprDeviceParameter
import redvypr.files as redvypr_files
import redvypr.gui
import redvypr.data_packets
from redvypr.widgets.pydanticConfigWidget import pydanticConfigWidget
from redvypr.gui import datastreamWidget
from redvypr.devices.plot import XYplotWidget
from redvypr.devices.plot import plot_widgets
from .calibration_models import calibration_HF, calibration_NTC, calibration_poly

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('autocalibration')
logger.setLevel(logging.DEBUG)

class AutoCalEntry(pydantic.BaseModel):
    """
    An entry for the autocalibration procedure
    """
    autocalmode: typing.Literal['timer', 'response', 'threshold'] = pydantic.Field(default='response',
                                                                                   description='The mode, i.e. the way autocal shall behave after the value is set.')
    parameter: RedvyprAddress = pydantic.Field(default=RedvyprAddress(),
                                                       description='The parameter to be changed')
    parameter_set: float = pydantic.Field(default=0.0, description='The value the parameter shall be changed to')
    parameter_steady: RedvyprAddress = pydantic.Field(default=RedvyprAddress(),
                                               description='The parameter giving the signal if the desired value is steady')
    parameter_steady_true: typing.Any = pydantic.Field(default=1, description='The value that is sent by parameter_steady when the value is steady')
    parameter_steady_false: typing.Any = pydantic.Field(default=0,
                                                   description='The value that is sent by parameter_steady when the value ist not yet steady')
    command: str = pydantic.Field(default='', description='The command to be sent')
    timer_wait: float = pydantic.Field(default=120, description='Seconds to wait')
    sample_delay: float = pydantic.Field(default=2.0, description='The delay [s] after which the paremeters are sampled')
    entry_min_runtime: float = pydantic.Field(default=5.0, description='The mininum time[s] an entry should run before analysis is performed')


class AutoCalConfig(AutoCalEntry):
    entries: typing.Optional[typing.List[AutoCalEntry]] = pydantic.Field(default=[], editable=True)
    start_index: int = pydantic.Field(default=0, description='The index of the entries to start with')
    parameter_delta: float = pydantic.Field(default=1.0,
                                            description='The delta between the next entry if a new is added',
                                            editable=True)
    parameter_start: float = pydantic.Field(default=0, description='The start value of the parameter',
                                            editable=True)
    autocalmode: typing.Literal['timer', 'response', 'threshold'] = pydantic.Field(default='timer',
                                                                                   description='The mode, i.e. the way autocal shall behave after the value is set.')


class autocalWidget(QtWidgets.QWidget):
    def __init__(self, device=None):
        """
        Standard deviceinitwidget if the device is not providing one by itself.

        Args:
            device:
        """
        funcname = __name__ + '.__init__():'
        logger.debug(funcname)
        super().__init__()
        self.device = device
        self.config = self.device.custom_config.autocal_config
        self.layout = QtWidgets.QGridLayout(self)
        self.calentrytable = QtWidgets.QTableWidget()
        self.parameterinput = QtWidgets.QPushButton('Parameter')
        self.parameterinput.clicked.connect(self.parameterClicked)
        self.parameterinput_edit = QtWidgets.QLineEdit()
        self.parameterinput_edit.setText(self.config.parameter.get_str())

        self.parameter_steady_edit = QtWidgets.QLineEdit()
        self.parameter_steady_edit.setText(self.config.parameter_steady.get_str())
        self.parameter_steady_button = QtWidgets.QPushButton('Parameter steady')
        self.parameter_steady_button.clicked.connect(self.parameterClicked)

        self.parameter_start = QtWidgets.QDoubleSpinBox()
        self.parameter_start.setMinimum(-1e10)
        self.parameter_start.setMaximum(1e10)
        self.parameter_start.setValue(self.config.parameter_start)

        self.parameter_delta = QtWidgets.QDoubleSpinBox()
        self.parameter_delta.setMinimum(-1e10)
        self.parameter_delta.setMaximum(1e10)
        self.parameter_delta.setValue(self.config.parameter_delta)

        self.addentry = QtWidgets.QPushButton('Add entry')
        self.addentry.clicked.connect(self.additem_clicked)
        self.rementry = QtWidgets.QPushButton('Remove entry')
        self.rementry.clicked.connect(self.remitem_clicked)
        self.start = QtWidgets.QPushButton('Start autocalibration')
        self.start.clicked.connect(self.start_clicked)
        self.start.setCheckable(True)
        self.start_index = QtWidgets.QSpinBox()
        self.start_index.setValue(self.config.start_index)
        self.start_index.valueChanged.connect(self.start_index_changed)
        #timer_wait
        self.parameter_steady_true = QtWidgets.QLineEdit()
        self.parameter_steady_true.setText(str(self.config.parameter_steady_true))
        self.parameter_steady_false = QtWidgets.QLineEdit()
        self.parameter_steady_false.setText(str(self.config.parameter_steady_false))
        self.timer_wait = QtWidgets.QDoubleSpinBox()
        self.timer_wait.setMinimum(0.0)
        self.timer_wait.setMaximum(10000.0)
        self.timer_wait.setValue(self.config.timer_wait)
        self.parameter_threshold = QtWidgets.QDoubleSpinBox()
        self.modecombo = QtWidgets.QComboBox()
        modes = typing.get_args(typing.get_type_hints(AutoCalConfig())['autocalmode'])
        for m in modes:
            self.modecombo.addItem(str(m))

        self.modecombo.setCurrentIndex(1)
        self.layout_config0 = QtWidgets.QHBoxLayout()
        self.layout_config0.addWidget(self.parameterinput)
        self.layout_config0.addWidget(self.parameterinput_edit)
        self.layout_config0.addWidget(QtWidgets.QLabel('Parameter start'))
        self.layout_config0.addWidget(self.parameter_start)
        self.layout_config0.addWidget(QtWidgets.QLabel('Parameter delta'))
        self.layout_config0.addWidget(self.parameter_delta)
        self.layout.addLayout(self.layout_config0, 0, 0,1,-1)
        self.layout_config1 = QtWidgets.QHBoxLayout()
        self.layout_config1.addWidget(QtWidgets.QLabel('Mode'))
        self.layout_config1.addWidget(self.modecombo)
        self.layout_config1.addWidget(self.parameter_steady_button)
        self.layout_config1.addWidget(self.parameter_steady_edit,3)
        self.layout_config1.addWidget(QtWidgets.QLabel('Parameter steady true'))
        self.layout_config1.addWidget(self.parameter_steady_true,1)
        self.layout_config1.addWidget(QtWidgets.QLabel('Parameter steady false'))
        self.layout_config1.addWidget(self.parameter_steady_false,1)
        self.layout_config1.addWidget(QtWidgets.QLabel('Timer wait'))
        self.layout_config1.addWidget(self.timer_wait)
        self.layout_config1.addWidget(QtWidgets.QLabel('Threshold'))
        self.layout_config1.addWidget(self.parameter_threshold)
        self.layout.addLayout(self.layout_config1, 1, 0, 1,-1)

        irow = 1
        self.layout.addWidget(self.addentry, 1+irow, 0,1,-1)
        self.layout.addWidget(self.calentrytable, 2+irow, 0, 1, -1)
        self.layout.addWidget(self.rementry, 3+irow, 0,1,-1)
        self.layout.addWidget(self.start, 4+irow, 0, 1, -1)
        self.layout.addWidget(QtWidgets.QLabel('Start index'), 4+irow, 2)
        self.layout.addWidget(self.start_index, 4+irow, 3)
        self.layout.addWidget(self.start, 4+irow, 0, 1, 2)

        self.col_status = 0
        self.col_addr = 1
        self.col_parameter_set = 2
        self.col_mode = 3
        self.col_timer_wait = 4
        self.col_addr_steady = 5
        self.col_sample_delay = 6
        self.col_value = 7

        self.colheader = ['Status', 'Address', 'Parameter set', 'Mode', 'Timer wait', 'Addr steady', 'Sample delay', 'Value']
        self.update_entrytable()
        # Adding myself to the guiqueue to get the data for autocalibration
        self.device.add_guiqueue(widget=self)

    def start_index_changed(self):
        funcname = __name__ + '.start_index_changed():'
        logger.debug(funcname)
        self.config.start_index = self.start_index.value()

    def start_clicked(self):
        funcname = __name__ + '.start_clicked():'
        logger.debug(funcname)
        if self.start.isChecked():
            if self.device.isRunning() == False:
                logger.debug(funcname + 'Calibration device is not running yet, starting first')
                self.device.thread_start()
            start_index = self.config.start_index
            self._autocal_entry_running = False
            logger.debug('Starting at index {}'.format(start_index))
            self._autocal_entry_index = start_index
            self.autcal_run_timer = QtCore.QTimer()
            self.autcal_run_timer.timeout.connect(self.autocal_run)
            self.autcal_run_timer.start(1000)
            self.start_index.setEnabled(False)
            self.autocal_run()
            self.start.setText('Stop autocalibration')
        else:
            logger.debug('Stopping')
            self.start_index.setEnabled(True)
            self.autcal_run_timer.stop()
            self.start.setText('Start autocalibration')

    def autocal_run(self):
        funcname = __name__ + '.autocal_run():'
        logger.debug(funcname)
        try:
            self._autocal_entry_run = self.config.entries[self._autocal_entry_index]
        except:
            logger.debug(funcname + 'Could not get entry, stopping')
            self.autcal_run_timer.stop()
            self.start.setChecked(False)
            self.start_index.setEnabled(True)
            self.start.setText('Start autocalibration')
            for irow, entry in enumerate(self.config.entries):
                item_status = QtWidgets.QTableWidgetItem('Idle')
                item_status.setFlags(item_status.flags() ^ QtCore.Qt.ItemIsEditable)
                self.calentrytable.setItem(irow, self.col_status, item_status)

            return

        self.start_index.setEnabled(False)
        #print('Processing entry',self._autocal_entry_index)
        # Check if an entry is processed
        parameter = self._autocal_entry_run.parameter
        #print('Device',parameter.devicename)
        #print('Datakey', parameter.datakey)


        if self._autocal_entry_running:
            #print('Running, doing nothing, could update status here')

            col_status = 0
            item_status = self.calentrytable.item(self.config.start_index, self.col_status)
            trun = time.time() - self._autocal_entry_tstart
            item_status.setText('Running {:.1f}'.format(trun))
            if self._autocal_entry_run.autocalmode == 'response':
                pass
                #print('Response')
        else:
            # Find the device
            devices = self.device.redvypr.get_device_objects()
            flag_device_found = False
            self._autocal_entry_device = None
            for d in devices:
                #print('d',d)
                if d.name == parameter.devicename:
                    logger.debug(funcname + 'Found a device')
                    flag_device_found = True
                    self._autocal_entry_device = d

            if self._autocal_entry_device is not None:
                logger.debug(funcname + 'Sending command')
                comdata = {'temp':self._autocal_entry_run.parameter_set}
                self._autocal_entry_device.thread_command(command='set', data=comdata)

            if self._autocal_entry_run.autocalmode == 'response':
                logger.debug(funcname + 'Processing entry with response mode')
                self.device.subscribe_address(self._autocal_entry_run.parameter)
                self.device.subscribe_address(self._autocal_entry_run.parameter_steady)
                self._autocal_response_steady = False
                self._autocal_entry_running = True
                self._autocal_entry_tstart = time.time()
            elif self._autocal_entry_run.autocalmode == 'timer':
                logger.debug(funcname + 'Processing entry with timer mode')
                self.device.subscribe_address(self._autocal_entry_run.parameter)
                self._autocal_entry_running = True
                self._autocal_entry_tstart = time.time()
                dt_wait = self._autocal_entry_run.timer_wait

                logger.debug(funcname + 'Waiting for {}'.format(dt_wait))
                self.autcal_run_timer_entry = QtCore.QTimer()
                self.autcal_run_timer_entry.timeout.connect(self.autocal_run_entry_wait)
                self.autcal_run_timer_entry.start(int(dt_wait * 1000))
                # Set item status
                item_status = self.calentrytable.item(self.config.start_index, self.col_status)
                trun = time.time() - self._autocal_entry_tstart
                item_status.setText('Running {:.1f}'.format(trun))

    def autocal_run_entry_wait(self):
        funcname = __name__ + '.autocal_run_entry_wait():'
        logger.debug(funcname + 'Waited long enough, sampling and increasing index')
        self.autcal_run_timer_entry.stop()
        self.autocal_run_next_entry()

    def autocal_run_next_entry(self):
        funcname = __name__ + '.autocal_run_next_entry():'
        # Increasing index
        # Status of table
        item_status = self.calentrytable.item(self.config.start_index, self.col_status)
        trun = time.time() - self._autocal_entry_tstart
        item_status.setText('Idle')
        self.config.start_index += 1
        start_index = self.config.start_index
        logger.debug('Index {}'.format(start_index))
        self._autocal_entry_index = start_index
        self.start_index.setValue(self.config.start_index)
        self._autocal_entry_running = False
        # Sampling
        logger.debug(funcname + 'Sampling')
        self.device.devicedisplaywidget.get_intervalldata()
        self.autocal_run() # Start a new round

    def update_entrytable(self):
        self.calentrytable.clear()

        self.calentrytable.setColumnCount(len(self.colheader))

        nrows = len(self.config.entries)
        self.calentrytable.setRowCount(nrows)

        self.calentrytable.setHorizontalHeaderLabels(self.colheader)
        for irow, entry in enumerate(self.config.entries):
            #print('Entry',entry)
            item_status = QtWidgets.QTableWidgetItem('Idle')
            item_status.setFlags(item_status.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_status, item_status)

            addr = entry.parameter.get_str('/d/k')
            item_addr = QtWidgets.QTableWidgetItem(addr)
            item_addr.setFlags(item_addr.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_addr, item_addr)

            parameter_set = str(entry.parameter_set)
            item_set = QtWidgets.QTableWidgetItem(parameter_set)
            item_set.setFlags(item_set.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_parameter_set,item_set)

            autocalmode = str(entry.autocalmode)
            item_mode = QtWidgets.QTableWidgetItem(autocalmode)
            item_mode.setFlags(item_mode.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_mode, item_mode)

            timer_wait = str(entry.timer_wait)
            item_wait = QtWidgets.QTableWidgetItem(timer_wait)
            item_wait.setFlags(item_wait.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_timer_wait, item_wait)

            sample_delay = str(entry.sample_delay)
            item_sample_delay = QtWidgets.QTableWidgetItem(sample_delay)
            item_sample_delay.setFlags(item_sample_delay.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_sample_delay, item_sample_delay)

            item_value = QtWidgets.QTableWidgetItem('NA')
            item_value.setFlags(item_value.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_value, item_value)

            addr_steady = entry.parameter_steady.get_str('/d/k')
            item_addr_steady = QtWidgets.QTableWidgetItem(addr_steady)
            item_addr_steady.setFlags(item_addr_steady.flags() ^ QtCore.Qt.ItemIsEditable)
            self.calentrytable.setItem(irow, self.col_addr_steady, item_addr_steady)

    def additem_clicked(self):
        funcname = __name__ + '.additem_clicked()'
        logger.debug(funcname)
        parameter_start = self.parameter_start.value()
        parameter_delta = self.parameter_delta.value()
        timer_wait = self.timer_wait.value()
        autocalmode = self.modecombo.currentText()
        self.parameter_start.setValue(parameter_start + parameter_delta)
        entry = AutoCalEntry(parameter_set=parameter_start, parameter=self.config.parameter,autocalmode=autocalmode,
                             parameter_steady=self.config.parameter_steady,
                             timer_wait=timer_wait)
        self.config.entries.append(entry)
        self.update_entrytable()

    def remitem_clicked(self):
        funcname = __name__ + '.remitem_clicked()'
        logger.debug(funcname)
        indexes = []
        for selectionRange in self.calentrytable.selectedRanges():
            indexes.extend(range(selectionRange.topRow(), selectionRange.bottomRow()+1))

        indexes.reverse()
        logger.debug(funcname + "Removing  indexes {}".format(indexes)) #
        for i in indexes:
            self.config.entries.pop(i)

        self.update_entrytable()

    def send_commnd_to_device(self,calentry):
        funcname = __name__ + '.send_commnd_to_device()'
        logger.debug(funcname)
        devicename = calentry.parameter.devicename
        #print('Devicename',devicename)

    def parameterClicked(self):
        self.pydantic_config = datastreamWidget(redvypr=self.device.redvypr)
        self.pydantic_config.apply.connect(self.parameter_changed)
        self.__pydantic_config_sender__ = self.sender()
        self.pydantic_config.show()
    def parameter_changed(self,config):
        funcname = __name__ + '.parameter_changed():'
        logger.debug(funcname + 'Config {}'.format(config))
        device = config['device']
        address = config['datastream_address']
        if self.__pydantic_config_sender__ == self.parameterinput:
            self.config.parameter = address
            self.parameterinput_edit.setText(self.config.parameter.get_str())
        elif self.__pydantic_config_sender__ == self.parameter_steady_button:
            self.config.parameter_steady = address
            self.parameter_steady_edit.setText(self.config.parameter_steady.get_str())

    def update_data(self,data):
        funcname = __name__ + '.update_data():'
        logger.debug(funcname)
        if data in self._autocal_entry_run.parameter:
            #print('Data for autocalibration',data)
            item_value = self.calentrytable.item(self.config.start_index, self.col_value)
            rdata = redvypr.data_packets.Datapacket(data)
            valuedata = rdata[self._autocal_entry_run.parameter]
            valuedatastr = str(valuedata)
            item_value.setText(valuedatastr)

        if self._autocal_entry_run.autocalmode == 'response':
            if data in self._autocal_entry_run.parameter_steady:
                trun = time.time() - self._autocal_entry_tstart
                #print('Found steady paramter')
                steady_true = self._autocal_entry_run.parameter_steady_true
                steady_false = self._autocal_entry_run.parameter_steady_false
                rdata = redvypr.data_packets.Datapacket(data)
                steadydata = rdata[self._autocal_entry_run.parameter_steady]
                #print('Steadydata',steadydata)
                # Check if steady and at least next_entry_min_time seconds

                if steadydata == steady_true and (trun > self._autocal_entry_run.entry_min_runtime):
                    if self._autocal_response_steady == False:
                        print('Parameter steady ...')
                        self._autocal_response_steady = True
                        self._autocal_response_steady_time = time.time()
                        #dtwaititem = self.calentrytable.item(self.config.start_index, self.col_timer_wait)
                        #dt_wait = float(dtwaititem.text())
                        #dtwaititem.dt_wait = dt_wait
                    else:
                        pass

                if self._autocal_response_steady:
                    print('Steady')
                    try:
                        dt = time.time() - self._autocal_response_steady_time
                        dtwaititem = self.calentrytable.item(self.config.start_index, self.col_timer_wait)
                        #dt_wait = int(dtwaititem.text())
                        dt_wait = self.config.entries[self.config.start_index].timer_wait
                        #dt_wait = dtwaititem.dt_wait
                        dt_left = dt_wait - dt
                        dtwaititem.setText('{:.2f}'.format(dt_left))
                        print('dt', dt, 'dt_wait', dt_wait)
                        if dt_left > 0:
                            print('Still waiting')
                            pass
                        else:
                            print('New round')
                            self._autocal_entry_running = False
                            self.autocal_run_next_entry()
                    except:
                        logger.warning('Could not process',exc_info=True)